# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
Created on Aug 27, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import pytest
from hope._transformer import ASTTransformer
from hope._ast import Module
from hope._dump import Dumper
import hope

def outer_dummy(x):
    return x

class DummyWrapper(object):
    a_a = 0
    a_b = np.array([1])
    def dummy(self, a):
        
        b=np.exp(a)
        c=np.zeros(10, dtype=np.float32)
        c[:] = b + a
        for i in range(3,10):
            b += i
        
        a[1:9] = -b[1:9]
        
        if 1 == 1 and True:
            a[:] = self.a_b[0]
        else:
            a[:] = self.a_a
            outer_dummy(a)
        
        d = True
        while d != True:
            d = False
        e = np.sum(b)
        return hope.exp(e)


@pytest.mark.infrastructure
class TestDumper(object):

    def test_dump_complex(self):
        wrapper = DummyWrapper()
        module = Module(wrapper.dummy.__name__)
        args = [wrapper, np.arange(10)]
        ASTTransformer(module).module_visit(wrapper.dummy, args)
        fkt_str = Dumper().visit(module)
        assert fkt_str is not None
        
        
        



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate the support for calling other functions

http://pythonhosted.org/hope/lang.html#call-functions

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil

import hope

prefix = "hope_calls"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

def fkt1(d):
    return d + 1 

@hope.jit
def demo1(d):
    s = fkt1(d)
    return s

@hope.jit
def fkt2(d):
    return d + 1 

@hope.jit
def demo2(d):
    s = fkt2(d)
    return s

@hope.jit
def fkt_recursion(n):
    if n < 2:
        return n
    return fkt_recursion(n - 1) + fkt_recursion(n - 2)

def example():
    #jitted function calls non-decorated function
    r = demo1(2)
    assert r == 3
    
    fkt2(0)
    #jitted function calls decorated function
    r = demo2(2)
    assert r == 3
    
    #recursive calls of a jitted function
    assert fkt_recursion(20) == 6765

if __name__ == '__main__':
    example()


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test functions for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import hope
import numpy as np

from .utilities import check,setup_module,setup_method,teardown_module


class SubCls(object):
    def __init__(self):
        self.i = 2


class Cls(object):
    def __init__(self):
        self.i = 1
        self.f = 1.
        self.ai = np.array([1, 2, 3], dtype=np.uint32)
        self.af = np.array([1, 2, 3], dtype=np.float32)
        self.res = np.array([0, 0, 0], dtype=np.float32)
        self.obj = SubCls()

    def fkt_cb(self):
        return self.i

    # @hope.jit
    # def fkt_jit(self):
    #   return self.f

    @hope.jit
    def fkt_1(self):
        self.res[:] = self.af

    @hope.jit
    def fkt_2(self):
        return self.fkt_cb()

    # @hope.jit
    # def fkt_3(self):
    #     return self.fkt_jit()

    @hope.jit
    def fkt_4(self):
        return self.obj.i


def test_cls_1():
    inst = Cls()
    inst.fkt_1()
    assert check(inst.af, inst.res)
    inst.af = np.array([1, 2, 3], dtype=np.int_)
    inst.fkt_1()
    assert check(inst.af, inst.res)


def test_cls_2():
    inst = Cls()
    assert check(inst.fkt_2(), inst.i)

# def test_cls_3():
#   inst = Cls()
#   assert check(inst.fkt_3(), inst.f)


def test_cls_4():
    inst = Cls()
    assert check(inst.fkt_4(), inst.obj.i)

def test_member_reference_array():
    class Test(object):
        def __init__(self, n=10):
            self.member = np.ones(n)

        def fkt(self):
            arr = self.member
            return arr
    
        hfkt = hope.jit(fkt)
            
    t1 = Test()
    arr = t1.fkt()
    harr = t1.hfkt()
    assert np.all(arr == harr)
    assert t1.member is harr

def test_member_reference_scalar():
    class Test(object):
        def __init__(self, n=10):
            self.member = n

        def fkt(self):
            a = self.member
            return a
    
        hfkt = hope.jit(fkt)
            
    t1 = Test()
    a = t1.fkt()
    ha = t1.hfkt()
    assert a == ha

def test_member_reference_view():
    class Test(object):
        def __init__(self, n=10):
            self.member = np.ones(n)
            self.out = np.empty(n)

        def fkt(self):
            self.out[:] = self.member
    
        hfkt = hope.jit(fkt)
            
    t1 = Test()
    t1.hfkt()
    assert t1.member is not t1.out
    assert np.all(t1.member == t1.out)

def test_member_reference_member():
    class Test(object):
        def __init__(self, n=10):
            self.member = np.ones(n)
            self.out = np.empty(n)

        def fkt(self):
            self.out = self.member
    
        hfkt = hope.jit(fkt)
            
    t1 = Test()
    t1.fkt()
    assert t1.member is t1.out

def test_member_operation():
    class Test(object):
        def __init__(self):
            self.member = np.ones(3)

        def fkt(self):
            return self.member + 1.0
    
        hfkt = hope.jit(fkt)
            
    t1 = Test()
    a = t1.fkt()
    ah = t1.hfkt()
    assert np.all(a==ah)



# encoding: utf-8
from __future__ import print_function, division, absolute_import

# we set a seed, because the legacy hope test code uses random data in somep places which caused
# unforeseeable failures when numpy changed its behaviour handling exponentiation of integer
# types.
import random
import numpy as np

random.seed(12345)
np.random.seed(12345)

import tempfile
import sys

import hope.config as config
config.prefix = tempfile.mkdtemp("hope")




# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate loop support

http://pythonhosted.org/hope/lang.html#loops

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil

import hope

prefix = "hope_loops"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def while_loop(d):
    while d > 0:
        d = d -1
    return d 

@hope.jit
def range_end(d):
    s = 0   
    for i in range(d):
        s = s + i
    return s

@hope.jit
def range_start_end(d):
    s = 0   
    for i in range(d, d*2):
        s = s + i
    return s


def example():
    r = while_loop(2)
    assert r == 0

    #equivalent support for xrange
    r = range_end(5)
    assert r == 10
    r = range_start_end(5)
    assert r == 35

if __name__ == '__main__':
    example()


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Created on Mar 12, 2014

author: jakeret
"""
from __future__ import print_function, division, absolute_import, unicode_literals


class UnsupportedCompilerException(Exception):
    pass

class UnsupportedFeatureException(Exception):
    pass


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test functions for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import itertools
import pytest
import copy

from .utilities import random, check, make_test, dtypes, shapes

import hope
from hope.exceptions import UnsupportedFeatureException

# TODO: fix for np.float32
@pytest.mark.parametrize("dtype", [np.float64, float])
def test_func_interp(dtype):
    
    def fkt(x, y, x0, y0, s0, l0, r0):
        y0[:] = np.interp(x0, x, y)
        l0[:] = np.interp(x0 - 1, x, y, left=-1)
        r0[:] = np.interp(x0 + 1, x, y, right=2)
        s0[0] = np.interp(x0[0], x, y)
    hfkt = hope.jit(fkt)
    x0 = np.linspace(0, 1, 50)
    x = np.linspace(0, 1, 10).astype(dtype)
    y = np.linspace(0, 1, 10).astype(dtype)
    xo, xh = np.linspace(0, 1, 50).astype(dtype), np.linspace(0, 1, 50).astype(dtype)
    yo, yh = np.zeros_like(x0), np.zeros_like(x0)
    so, sh = np.zeros_like(x0), np.zeros_like(x0)
    lo, lh = np.zeros_like(x0), np.zeros_like(x0)
    ro, rh = np.zeros_like(x0), np.zeros_like(x0)
    fkt(x, y, xo, yo, so, lo, ro),  hfkt(x, y, xh, yh, sh, lh, rh)
    assert check(xo, xh)
    assert check(yo, yh)
    assert check(so, sh)
    assert check(lo, lh)
    assert check(ro, rh)
    
@pytest.mark.parametrize("dtype", [np.float64, float])
def test_func_interp_bounds(dtype):
    def fkt(x,y,x0,y0):
        y0[:] = np.interp(x0, x, y)
    
    hfkt = hope.jit(fkt)
    x = np.arange(5,15).astype(dtype)
    y = -x
    x0 = np.arange(0, 20).astype(dtype)
    yo, yh = np.zeros_like(x0), np.zeros_like(x0)
    fkt(x, y, x0, yo), hfkt(x, y, x0, yh)
    assert check(yo, yh)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_sin(a, b, c): return np.sin(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_cos(a, b, c): return np.cos(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_tan(a, b, c): return np.tan(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_sign(a, b, c): return np.sign(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
def test_func_arcsin(dtype, shape):
    def fkt(a): 
        return np.arcsin(a)
    
    hfkt = hope.jit(fkt)
    a = 2 * np.random.random(shape) - 1
    b = copy.deepcopy(a)
    co = fkt(a)
    ch = hfkt(b)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
def test_func_arccos(dtype, shape):
    def fkt(a): 
        return np.arccos(a)
    
    hfkt = hope.jit(fkt)
    a = 2 * np.random.random(shape) - 1
    b = copy.deepcopy(a)
    co = fkt(a)
    ch = hfkt(b)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_arctan(a, b, c): return np.arctan(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
def test_func_sinh(dtype, shape):
    def fkt(a): 
        return np.sinh(a)
    
    hfkt = hope.jit(fkt)
    a = np.log(np.abs(random(dtype, shape)[0]) / 2.)
    b = copy.deepcopy(a)
    co = fkt(a)
    ch = hfkt(b)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
def test_func_cosh(dtype, shape):
    def fkt(a): 
        return np.cosh(a)
    
    hfkt = hope.jit(fkt)
    a = np.log(np.abs(random(dtype, shape)[0]) / 2.)
    b = copy.deepcopy(a)
    co = fkt(a)
    ch = hfkt(b)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_tanh(a, b, c): return np.tanh(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_floor(a, b, c): return np.floor(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_ceil(a, b, c): return np.ceil(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_trunc(a, b, c): return np.trunc(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
def test_func_sqrt(dtype, shape):
    def fkt(a): 
        return np.sqrt(a)
    
    hfkt = hope.jit(fkt)
    (ao, ah) = random(dtype, shape)
    ao, ah = np.abs(ao), np.abs(ah)
    co = fkt(ao)
    ch = hfkt(ah)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
@make_test
def test_func_fabs(a, b, c): return np.fabs(a)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
def test_func_exp(dtype, shape):
    def fkt(a): 
        return np.exp(a)
    
    hfkt = hope.jit(fkt)
    ao, ah = np.log(np.abs(random(dtype, shape))).astype(dtype)
    co = fkt(ao)
    ch = hfkt(ah)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product([np.float32, np.float64, float], shapes))
def test_func_log(dtype, shape):
    def fkt(a): 
        return np.log(a)
    
    hfkt = hope.jit(fkt)
    a, b = np.abs(random(dtype, shape))
    co = fkt(a)
    ch = hfkt(b)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
def test_func_sum_var(dtype, shape):
    def fkt(a):
        return np.sum(a)
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, shape)
    ao, ah = ao / 1200, ah / 1200
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
def test_func_sum_expr(dtype, shape):
    def fkt(a, b):
        return np.sum(a + b)
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh) = random(dtype, shape), random(dtype, shape)
    ao, ah = np.abs(ao) / 2400, np.abs(ah) / 2400
    bo, bh = np.abs(bo) / 2400, np.abs(bh) / 2400
    co, ch = fkt(ao, bo), hfkt(ah, bh)
    assert check(co, ch)
    co, ch = fkt(ao, bo), hfkt(ah, bh)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
def test_func_sum_dtype(dtype, shape):
    def fkt(a):
        return np.sum(a, dtype=np.float64)
    
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, shape)
    ao, ah = ao / 1200, ah / 1200
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)
    assert co.dtype == np.float64
    assert ch.dtype == np.float64

def test_func_sum_invalid_args():
    a = np.random.random(1)
    b = np.empty_like(a)
    
    def fkt(a, b):
        return np.sum(a, b)
    
    hfkt = hope.jit(fkt)
    with pytest.raises(UnsupportedFeatureException):
        hfkt(a, b)

    def fkt_axis(a):
        return np.sum(a, axis=0)
    
    hfkt = hope.jit(fkt_axis)
    with pytest.raises(UnsupportedFeatureException):
        hfkt(a)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes))
def test_func_sum_in_if(dtype, shape):
    def fkt(a):
        if True:
            val = np.sum(a)
        else:
            val = 1
        return val
    hfkt = hope.jit(fkt)
    ao, ah = random(dtype, shape)
    ao, ah = ao / 1200, ah / 1200
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)

def test_create_empty_array():
    def fkt(shape):
        return np.empty(shape)
        
    hfkt = hope.jit(fkt)
    assert fkt(5).size == hfkt(5).size

    def fkt_dtype(shape):
        return np.empty(shape, dtype=np.float64)
        
    hfkt = hope.jit(fkt)
    co, ch = fkt(5), hfkt(5)
    assert co.size == ch.size
    assert co.dtype == np.float64
    assert ch.dtype == np.float64

def test_create_zeros_array():
    def fkt(shape):
        return np.zeros(shape)
        
    hfkt = hope.jit(fkt)
    assert fkt(5).size == hfkt(5).size

    def fkt_dtype(shape):
        return np.zeros(shape, dtype=np.float64)
        
    hfkt = hope.jit(fkt)
    co, ch = fkt(5), hfkt(5)
    assert co.size == ch.size
    assert co.dtype == np.float64
    assert ch.dtype == np.float64

def test_create_ones_array():
    def fkt(shape):
        return np.ones(shape)
        
    hfkt = hope.jit(fkt)
    assert fkt(5).size == hfkt(5).size

    def fkt_dtype(shape):
        return np.ones(shape, dtype=np.float64)
        
    hfkt = hope.jit(fkt_dtype)
    co, ch = fkt(5), hfkt(5)
    assert co.size == ch.size
    assert co.dtype == np.float64
    assert ch.dtype == np.float64

def test_create_array_invalid_args():
    def fkt(shape):
        return np.empty(shape, 0)
        
    hfkt = hope.jit(fkt)
    with pytest.raises(UnsupportedFeatureException):
        hfkt(5)

    def fkt_order(shape):
        return np.empty(shape, order="c")
        
    hfkt = hope.jit(fkt_order)
    with pytest.raises(NotImplementedError):
        hfkt(5)

def test_unsupported_np_func():
    def fkt(a):
        return np.alen(a)

    hfkt = hope.jit(fkt)
    a = np.random.random(1)
    
    with pytest.raises(UnsupportedFeatureException):
        hfkt(a)


def test_np_func_invalid_args():
    def fkt(a, b):
        return np.log(a, b)
    
    hfkt = hope.jit(fkt)
    a = np.random.random(1)
    b = np.empty_like(a)
    
    with pytest.raises(UnsupportedFeatureException):
        hfkt(a, b)
    
    def fkt_kwargs(a, b):
        return np.log(a, out=b)
    
    hfkt = hope.jit(fkt_kwargs)
    with pytest.raises(UnsupportedFeatureException):
        hfkt(a, b)
    


# TODO: add tests for remaining functions




# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate some of the supported data types

http://pythonhosted.org/hope/lang.html#native-types
http://pythonhosted.org/hope/lang.html#numpy-types

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil
import numpy as np

import hope

prefix = "hope_datatypes"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def demo(d):
    return d

def example():
    #built-in
    demo(1)
    demo(1.0)
     
    #numpy 1d array
    demo(np.array([1], dtype=np.int8))
    demo(np.array([1], dtype=np.uint8))
    demo(np.array([1], dtype=np.float32))
    
    #numpy 2d array
    demo(np.array([[1], [2]], dtype=np.int8))
    demo(np.array([[1], [2]], dtype=np.uint16))
    demo(np.array([[1], [2]], dtype=np.float32))
    
    
    
if __name__ == '__main__':
    example()


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate the supported functionality with NumPy arrays

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil
import numpy as np
import hope

prefix = "hope_arrays"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def create_empty(n):
    return np.empty(n)

@hope.jit
def array_slicing_1D():
    a = np.zeros(4)
    a[:] = 3 
    a[:2] = 1
    a[1:2] = 2
    a[3:] = 4
    return a

@hope.jit
def array_slicing_2D(a):
    a[:] = 3
    a[1:, :] = 1
    a[:, 1:] = 2
    a[2:, 2:] = 4
    a[3, 3] = 5
    a[1, :-2] = 6
    a[:-2, 1] = 7
    a[0, -1] = 8
    return a

@hope.jit
def array_asignment(a, b):
    a[2:] = b[:2]

def example():
    #create new arrays (same for ones and zeros). Only 1D
    r = create_empty(5)
    assert len(r) == 5
    
    #slicing 1D array to assign new values
    r = array_slicing_1D()
    print(r)
    assert sum(r) == 10
    
    #slicing 2D array to assign new values
    a = np.zeros((4,4))
    r = array_slicing_2D(a)
    print(r)
    assert np.sum(r) == 60
    
    #assign a slice of an array to a slice
    a = np.zeros(4)
    b = np.ones(4)
    array_asignment(a, b)
    print(a)
    assert np.sum(a) == 2
    
if __name__ == '__main__':
    example()


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Created on Mar 11, 2014

author: jakeret
"""

from __future__ import print_function, division, absolute_import, unicode_literals

from operator import eq
from mock import patch
from mock import MagicMock

from hope.options import get_cxxflags
from hope import options
import pytest
from hope.exceptions import UnsupportedCompilerException

DARWIN_COMPILERS = ["gcc-mac", "clang", "gcc", "cc", "c++", "icc"]
DARWIN_MAPPING = {"gcc": "gcc-mac",
                  "cc": "clang",
                  "c++": "clang",
                  }

LINUX_COMPILERS = ["gcc-linux", "gcc", "cc", "c++", "icc", "x86_64-linux-gnu-gcc"]
LINUX_MAPPING = {"gcc": "gcc-linux",
                  "cc": "gcc",
                  "c++": "gcc",
                  "x86_64-linux-gnu-gcc": "gcc-linux"
                  }

# custom marker enabling selective unit testing
# e.g. py.test -v -m "infrastructure"

@pytest.mark.infrastructure
class TestOptions(object):


    @pytest.mark.parametrize("compiler_name", DARWIN_COMPILERS)
    def test_get_cxxflags_Darwin(self, compiler_name):
        with patch("platform.system") as sys_mock:
            sys_mock.return_value = options.DARWIN_KEY
            with patch("distutils.ccompiler.new_compiler") as new_cc_mock:
                with patch("subprocess.check_output") as command_mock:
                    command_mock.return_value = options.MIN_GCC_VERSION.encode('utf-8')
                    cc_mock = MagicMock()
                    cc_mock.compiler = [compiler_name]
                    new_cc_mock.return_value = cc_mock
                    flags = get_cxxflags()
                    assert flags is not None
                    
                    try:
                        compiler = DARWIN_MAPPING[compiler_name]
                    except KeyError:
                        compiler = compiler_name
                    
                    assert map(eq, flags, options.CXX_FLAGS[compiler])

    @pytest.mark.parametrize("compiler_name", LINUX_COMPILERS)
    def test_get_cxxflags_Linux(self, compiler_name):
        with patch("platform.system") as sys_mock:
            sys_mock.return_value = options.LINUX_KEY
            with patch("distutils.ccompiler.new_compiler") as new_cc_mock:
                with patch("subprocess.check_output") as command_mock:
                    command_mock.return_value = options.MIN_GCC_VERSION.encode('utf-8')
                    cc_mock = MagicMock()
                    cc_mock.compiler = [compiler_name]
                    new_cc_mock.return_value = cc_mock
                    flags = get_cxxflags()
                    assert flags is not None
                    
                    try:
                        compiler = LINUX_MAPPING[compiler_name]
                    except KeyError:
                        compiler = compiler_name
                    
                    assert map(eq, flags, options.CXX_FLAGS[compiler])
                
    @pytest.mark.parametrize("compiler_name", ["gcc-linux", "gcc"]) 
    def test_get_cxxflags_unsupported_version(self, compiler_name):
        with patch("platform.system") as sys_mock:
            sys_mock.return_value = options.LINUX_KEY
            with patch("distutils.ccompiler.new_compiler") as new_cc_mock:
                with patch("subprocess.check_output") as command_mock:
                    command_mock.return_value = "1.0.0".encode('utf-8')
                    cc_mock = MagicMock()
                    cc_mock.compiler = [compiler_name]
                    new_cc_mock.return_value = cc_mock
                    
                    try:
                        get_cxxflags()
                        pytest.fail("Version should not be supported")
                    except UnsupportedCompilerException:
                        assert True
    
    def test_get_cxxflags_invalid_os(self):
        with patch("platform.system") as sys_mock:
            sys_mock.return_value = "Win32"
            
            try:
                flags = get_cxxflags()
                pytest.fail("OS is not supported")
            except UnsupportedCompilerException:
                assert True

    def test_get_cxxflags_invalid_compiler(self):
        with patch("platform.system") as sys_mock:
            sys_mock.return_value = options.LINUX_KEY
            with patch("distutils.ccompiler.new_compiler") as new_cc_mock:
                cc_mock = MagicMock()
                cc_mock.compiler = ["invalid_cc"]
                new_cc_mock.return_value = cc_mock
                try:
                    flags = get_cxxflags()
                    pytest.fail("Compiler is not supported")
                except UnsupportedCompilerException:
                    assert True

    def test_get_cxxflags_gcc_clang(self):
        compiler_name = "gcc"
        with patch("platform.system") as sys_mock:
            sys_mock.return_value = options.DARWIN_KEY
            with patch("distutils.ccompiler.new_compiler") as new_cc_mock:
                with patch("subprocess.check_output") as command_mock:
                    command_mock.return_value = options.GCC_CLANG_VERSION.encode('utf-8')
                    cc_mock = MagicMock()
                    cc_mock.compiler = [compiler_name]
                    new_cc_mock.return_value = cc_mock
                    flags = get_cxxflags()
                    assert flags is not None
                    
                    try:
                        compiler = DARWIN_MAPPING[compiler_name]
                    except KeyError:
                        compiler = compiler_name
                    
                    assert map(eq, flags, options.CXX_FLAGS[compiler])

if __name__ == '__main__':
    test = TestOptions()
    test.test_get_cxxflags_Darwin()



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
Created on Aug 25, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import pytest
from hope import config
import shutil
from hope import serialization
import os
import tempfile

@pytest.mark.infrastructure
class TestSerialization(object):
    
    def setup(self):
        self.prefix = config.prefix
        config.prefix = tempfile.mkdtemp()
        
    
    def test_serialization(self):
        name = "obj"
        obj = {"test":1}
        serialization.serialize(obj, name)
        assert os.path.join(config.prefix, "obj.pck")
        
        obj2 = serialization.unserialize(name)
        assert obj == obj2
        
    def test_serialize_inexistent_prefix(self):
        name = "obj"
        obj = {"test":1}
        prefix = config.prefix
        config.prefix = os.path.join(prefix, "dir")
        serialization.serialize(obj, name)
        assert os.path.exists(config.prefix)
        assert os.path.join(config.prefix, "obj.pck")
        config.prefix = prefix
    
    def test_unserialize_inexistent(self):
        name = "obj1"
        obj = serialization.unserialize(name)
        assert obj == None
    
    def tear_down(self):
        shutil.rmtree(config.prefix)
        config.prefix = self.prefix



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
Created on Sep 26, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals
import pytest
from hope import _wrapper
import hope
from hope.jit import _check_state, jit

def dummy():
    pass

def dummy_called():
    pass

@pytest.mark.infrastructure
class TestJit(object):
    
    @pytest.fixture
    def config_state(self):
        state = {}
        for name in _wrapper.get_config_attrs():
            state[name] = getattr(hope.config, name)
            
        state["filename"] = "filename"
        return state
    
    def test_inconsistent_config(self):
        state = {"cxxflags": "new_flag"}
        
        try:
            _check_state(None, state)
            pytest.fail("Config is inconsistent")
        except LookupError as le:
            assert True
            
    def test_no_fkt_in_state(self, config_state):
        try:
            _check_state(None, config_state)
            pytest.fail("Config is inconsistent")
        except LookupError as le:
            assert True
            
        config_state["main"] = "fkt_name"
        try:
            _check_state(None, config_state)
            pytest.fail("Config is inconsistent")
        except LookupError as le:
            assert True
            
        config_state["called"] = "called_fkt_name"
        try:
            _check_state(dummy, config_state)
            pytest.fail("Config is inconsistent")
        except LookupError as le:
            assert True
            
    def test_missing_calling_fkt(self, config_state):
        config_state["main"] = dummy.__name__
        config_state["called"] = {}
        _check_state(dummy, config_state)
        
        config_state["called"]["called1"] = "hash"
        try:
            _check_state(dummy, config_state)
            pytest.fail("Config is inconsistent")
        except LookupError as le:
            assert True

        config_state["called"].pop("called1")
        config_state["called"][dummy_called.__name__] = "hash"
        try:
            _check_state(dummy, config_state)
            pytest.fail("Config is inconsistent")
        except LookupError as le:
            assert True

        config_state["called"].pop(dummy_called.__name__)
        config_state["called"][_check_state.__name__] = "hash"
        try:
            _check_state(dummy, config_state)
            pytest.fail("Config is inconsistent")
        except LookupError as le:
            assert True

    def test_fkt_with_args(self):
        with pytest.raises(ValueError):
            @jit
            def func_with_star_args(a, *args):
                pass
        with pytest.raises(ValueError):
            @jit
            def func_with_kwargs(a, **kwargs):
                pass
        with pytest.raises(ValueError):
            @jit
            def func_with_star_args_and_kwargs(a, *args, **kwargs):
                pass
        # Should not raise ValueError
        @jit
        def func_with_vanilla_args(a, b):
            pass



# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>
"""
Contains all the HOPE AST nodes currently supported

"""

from __future__ import print_function, division, absolute_import, unicode_literals


import ast
import copy

from hope._const import *
from hope.exceptions import UnsupportedFeatureException

class Token(object):
    def getId(self):
        return "{0}{1!s}".format(PY_TYPE_CHAR[self.dtype], "" if len(self.shape) == 0 else len(self.shape))

    def __ne__(self, other):
        return not self == other
    
    
    def __str__(self):
        from hope._dump import Dumper
        return Dumper().visit(self)


class Number(Token):
    def __init__(self, value):
        if not type(value) in PY_C_TYPE:
            raise Exception("Numberformat not implemented: {0}.{1!s}".format(value, type(value)))
        self.value, self.shape, self.dtype = value, [], type(value)
    # TODO: implement cmp

    def __eq__(self, other):
        return isinstance(other, Number) and self.value == other.value and self.dtype == other.dtype


class NewVariable(Token):
    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        return isinstance(other, NewVariable) and self.name == other.name


class Variable(Token):
    def __init__(self, name, shape, dtype = None, scope = "block", allocated = False):
        self.name, self.shape, self.dtype, self.scope, self.allocated = name, shape, dtype, scope, allocated

    def __eq__(self, other):
        return isinstance(other, Variable) and self.name == other.name and self.dtype == other.dtype and self.shape == other.shape and self.scope == other.scope


class Object(Token):
    def __init__(self, name, instance, parent=None):
        self.parent, self.name, self.classname, self.instance, self.attrs = parent, name, instance.__class__.__name__, instance, {}

    def getId(self, prefix = "o"):
        chars = "{0}{1}_".format(prefix, self.classname)
        for key in sorted(self.attrs.keys()):
            if isinstance(self.attrs[key], Object):
                chars += self.attrs[key].getId()
            elif len(self.attrs[key].shape) == 0:
                chars += PY_TYPE_CHAR[self.attrs[key].dtype]
            else:
                chars += "{0}{1}".format(PY_TYPE_CHAR[self.attrs[key].dtype], len(self.attrs[key].shape))
        return chars

    def getAttr(self, name):
        if not name in self.attrs:
            value = getattr(self.instance, name)
            if isinstance(value, np.ndarray):
                if value.dtype.isbuiltin != 1:
                    raise Exception("Only Builtin datatypes are supported: in {0} value {1} has type {2!s}".format(self.name, name, value.dtype))
                self.attrs[name] = ObjectAttr(self, name, value.dtype.type, len(value.shape), PY_TYPE_CHAR[value.dtype.type])
            elif hasattr(value, "dtype"):
                self.attrs[name] = ObjectAttr(self, name, value.dtype.type, 0, PY_TYPE_CHAR[value.dtype.type])
            elif isinstance(value, int): 
                self.attrs[name] = ObjectAttr(self, name, int, 0, PY_TYPE_CHAR[int])
            elif isinstance(value, float):
                self.attrs[name] = ObjectAttr(self, name, float, 0, PY_TYPE_CHAR[float])
            else:
                self.attrs[name] = Object(name, value, self)
        return self.attrs[name]

    def __eq__(self, other):
        if (self.parent is None) != (other.parent is None) or (self.parent is not None and other.parent is not None and self.parent != other.parent):
            return False
        return isinstance(other, Object) and self.name == other.name and self.classname == other.classname and self.attrs == other.attrs

    # on deepcopy do not copy the parent object, since we have not information about ist and do not require it to be deepcopy
    # ths violates the python semantichs :(
    def __deepcopy__(self, memo):
        memo[id(self)] = Object.__new__(Object)
        for key, value in list(self.__dict__.items()):
            setattr(memo[id(self)], key, value if key == "instance" and value is not None else copy.deepcopy(value, memo))
        return memo[id(self)]


class ObjectAttr(Token):
    def __init__(self, parent, name, dtype, dims, char):
        self.parent, self.name, self.shape, self.dtype, self.char = parent, name, [(None, Dimension(self, dim)) for dim in range(dims)], dtype, char

    def __eq__(self, other):
        return isinstance(other, ObjectAttr) and self.parent == other.parent and self.name == other.name and self.dtype == other.dtype and self.shape == other.shape and self.char == other.char

    def getTrace(self):
        parent = self.parent
        trace = [self.name]
        while not parent is None:
            trace.insert(0, parent.name)
            parent = parent.parent
        return trace

class Dimension(Token):
    def __init__(self, variable, dim):
        self.variable, self.shape, self.dim = variable, [], dim

    def __eq__(self, other):
        return isinstance(other, Dimension) and self.dim == other.dim \
            and isinstance(self.variable, Variable) and self.variable.name == other.variable.name and self.variable.dtype == other.variable.dtype \
            and len(self.variable.shape) == len(other.variable.shape) and self.variable.scope == other.variable.scope

class DimensionSlice(Dimension):
    def __init__(self, variable, dim, slice):
        super(DimensionSlice, self).__init__(variable, dim)
        self.slice = slice

class View(Token):
    def __init__(self, variable, extents):
        if not isinstance(variable, (Variable, ObjectAttr)):
            raise Exception("A view can only be constructed on a variable or a objectAttr")
        if len(extents) == 0:
            raise Exception("A View needs to have an extent")
        if len(extents) == 1 and extents[0] == (None, None):
            extents = copy.deepcopy(variable.shape)
        if len(variable.shape) != len(extents):
            raise Exception("Extends of variable and subscript do not match")
        self.variable, self.dtype, self.extents, self.shape = variable, variable.dtype, extents, []
        for ind, (variable_extent, extent) in enumerate(zip(self.variable.shape, extents[:])):
            # TODO: check if variable extends and extends do match
            if isinstance(extent, tuple):
                lower, upper = extent
                if lower is None: lower = variable_extent[0]
                if isinstance(lower, Number) and lower.value == 0: lower = None
                if upper is None: upper = variable_extent[1]
                if isinstance(upper, Number) and upper.value < 0: 
                    upper = DimensionSlice(variable, variable_extent[1].dim, upper)
                    extents[ind] = (lower, copy.deepcopy(upper))
                self.shape.append((lower, upper))

    def __eq__(self, other):
        return isinstance(other, View) and self.variable == other.variable and self.dtype == other.dtype and self.shape == other.shape


class Assign(Token):
    def __init__(self, target, value):
        self.target, self.value, self.dtype, self.shape = target, value, target.dtype, target.shape

    def __eq__(self, other):
        return isinstance(other, Assign) and self.target == other.target and self.value == other.value


class AugAssign(Token):
    def __init__(self, op, target, value):
        self.op, self.target, self.value, self.dtype, self.shape = AUGMENTED_ASSIGN[type(op).__name__], target, value, target.dtype, target.shape

    def __eq__(self, other):
        return isinstance(other, AugAssign) and self.op == other.op and self.target == other.target and self.value == other.value


class Expr(Token):
    def __init__(self, value):
        self.value, self.shape = value, value.shape

    def __eq__(self, other):
        return isinstance(other, Expr) and self.value == other.value


class UnaryOp(Token):
    def __init__(self, op, operand):
        opname, dtypefkt = UNARY_OPERATORS[op]
        dtype = None if operand.dtype is None else dtypefkt(operand.dtype)
        self.op, self.operand, self.dtype, self.shape = opname, operand, dtype, operand.shape

    def __eq__(self, other):
        return isinstance(other, UnaryOp) and self.op == other.op and self.operand == other.operand


class BinOp(Token):
    def __init__(self, op, left, right):
        opname, dtypefkt = BINARY_OPERATORS[op]
        dtype = None if left.dtype is None or right.dtype is None else dtypefkt(left.dtype, right.dtype)
        shape = left.shape if len(left.shape) > 0 else right.shape
        self.op, self.left, self.right, self.dtype, self.shape = opname, left, right, dtype, shape

    def __eq__(self, other):
        return isinstance(other, BinOp) and self.op == other.op and self.left == other.left and self.right == other.right


class BoolOp(Token):
    def __init__(self, op, values):
        self.op, self.values, self.dtype, self.shape = BOOL_OPERATORS[op], values, np.bool_, values[0].shape

    def __eq__(self, other):
        return isinstance(other, BoolOp) and self.op == other.op and all([otherval == selfval for otherval, selfval in zip(other.values, self.values)])


class Compare(Token):
    def __init__(self, op, left, comparator):
        opname, dtypefkt = COMPARE_OPERATORS[type(op).__name__]
        dtype = None if left.dtype is None or comparator.dtype is None else dtypefkt(left.dtype, comparator.dtype)
        shape = left.shape if len(left.shape) > 0 else comparator.shape
        self.op, self.left, self.comparator, self.dtype, self.shape = opname, left, comparator, dtype, shape

    def __eq__(self, other):
        return isinstance(other, BinOp) and self.op == other.op and self.left == other.left and self.comparator == other.comparator


class If(Token):
    def __init__(self, test, body, orelse):
        self.test, self.dtype, self.shape, self.body, self.orelse = test, None, [], body, orelse
    def __eq__(self, other):
        return isinstance(other, If) and self.test == other.test and self.body == other.body and self.orelse == other.orelse


class For(Token):
    def __init__(self, iter, lower, upper, body):
        self.iter, self.dtype, self.shape, self.lower, self.upper, self.body = iter, None, [], lower, upper, body

    def __eq__(self, other):
        return isinstance(other, For) and self.iter == other.iter and self.lower == other.lower and self.upper == other.upper and self.body == other.body


class While(Token):
    def __init__(self, test, body):
        self.test, self.dtype, self.shape, self.body = test, np.bool_, [], body

    def __eq__(self, other):
        return isinstance(other, While) and self.test == other.test and self.body == other.body


class Call(Token):
    def __init__(self, name, iterable, args = [], keywords = {}):
        self.name, self.dtype, self.shape, self.args, self.keywords = name, name.get_dtype(args, keywords), name.get_shape(args, keywords) if iterable else [], args, keywords


# TODO: make this nicer, implement return values
class GlobalFunction(Token):
    # TODO: make argid token to change argid by adapting object
    def __init__(self, name, shape, dtype):
        self.name, self.shape, self.dtype = name, shape, dtype

    def get_dtype(self, args, keywords):
        return self.dtype

    def get_shape(self, args, keywords):
        return self.shape


# TODO: make this nicer
class HopeAttr(Token):
    def __init__(self, name):
        self.name, self.shape = name, []

    def __check(self, args, keywords):
        if self.name == "exp":
            if len(args) != 1 or len(keywords) != 0:
                raise Exception("hope.exp requires exaclty one argument")
        else:
            raise Exception("Unknown hope attribute: hope.pow {0!s}".format(self.name))

    def get_dtype(self, args, keywords):
        self.__check(args, keywords)
        if self.name == "exp":
            return args[0].dtype

    def get_shape(self, args, keywords):
        self.__check(args, keywords)
        if self.name == "exp":
            return args[0].shape


# TODO: make this nicer
class NumpyAttr(Token):
    def __init__(self, name):
        self.name, self.shape = name, []

    def __check(self, args, keywords):
        if self.name in ["empty", "zeros", "ones"]:
            if len(args) != 1:
                raise UnsupportedFeatureException("numpy.{0} requires one argument".format(self.name))
            if len(keywords) > 1:
                raise Exception("numpy.{0} requires at most one keyword argument: dtype".format(self.name))
            if len(keywords) == 1 and (
                                       "dtype" not in keywords or \
                                       not isinstance(keywords["dtype"], NumpyAttr) or \
                                       not hasattr(np, keywords["dtype"].name) or \
                                       not getattr(np, keywords["dtype"].name) in PY_C_TYPE):
                raise Exception("numpy.{0} requires a dtype".format(self.name))
            
        elif self.name == "interp":
            for keyword in list(keywords.keys()):
                if not keyword in ["left", "right"]:
                    raise Exception("numpy.interp only allows left and right as keywords")
            if len(args) != 3:
                raise Exception("numpy.interp requires three arguments")
            for idx in [1, 2]:
                if len(args[idx].shape) != 1:
                    raise Exception("numpy.interp requires the second and third argument to be a 1d array")
                
        elif self.name in NPY_UNARY_FUNCTIONS or self.name in NPY_CAST_FUNCTIONS:
            if len(args) != 1 or len(keywords) != 0:
                raise UnsupportedFeatureException("HOPE only supports one argument for 'numpy.%s'. Received: %s"%(self.name, len(args) + len(keywords)))
        else:
            raise UnsupportedFeatureException("Unknown numpy function: numpy.{0}".format(self.name))

    def get_dtype(self, args, keywords):
        self.__check(args, keywords)
        if self.name in ["empty", "zeros", "ones"]:
            return getattr(np, keywords["dtype"].name) if len(keywords) == 1 else np.float64
        elif self.name in ["interp", "sign"] or self.name in NPY_UNARY_FUNCTIONS:
            return args[0].dtype
        elif self.name in NPY_CAST_FUNCTIONS:
            return NPY_CAST_FUNCTIONS[self.name]

    def get_shape(self, args, keywords):
        self.__check(args, keywords)
        if self.name in ["empty", "zeros", "ones"]:
            return [(None, arg) for arg in args[0]] if isinstance(args[0], list) else [(None, args[0])]
        elif self.name in ["interp", "sign"] or self.name in NPY_UNARY_FUNCTIONS or self.name in NPY_CAST_FUNCTIONS:
            return args[0].shape


class NumpyContraction(Token):
    def __init__(self, variable, op, value):
        self.variable, self.op, self.value, self.shape, self.dtype = variable, op, value, [], variable.dtype


class Allocate(Token):
    def __init__(self, variable):
        if variable.dtype is None:
            raise Exception("Allocation requires dtype: {0!s}".format(variable))
        self.variable, self.shape = variable, []


class Return(Token):
    def __init__(self, value):
        self.value, self.shape = value, []


class Block(Token):
    def __init__(self, expr):
        self.body, self.shape, self.merged = [expr], expr.shape, None

class Reference(Token):
    def __init__(self, target, value):
        self.target, self.value, self.dtype, self.shape = target, value, target.dtype, []

    def __eq__(self, other):
        return isinstance(other, Reference) and self.target == other.target and self.value == other.value

class Body(Token):
    def __init__(self, blocks):
        self.blocks, self.dtype, self.shape = blocks, None, []


class FunctionDef(Token):
    def __init__(self, name, signature):
        self.name, self.signature, self.shape, self.dtype, self.merged, self.shapes, self.decl, self.optimized, self.isbound \
            = name, signature, None, None, {}, {}, None, False, len(signature) > 0 and isinstance(signature[0], Object)
        self.return_allocated = True

    def getId(self):
        return "".join([arg.getId() for arg in self.signature])


class Module(Token):
    def __init__(self, main):
        self.main, self.functions = main, {}


class NodeVisitor(object):
    def visit(self, node):
        if hasattr(self, 'visit_{0}'.format(type(node).__name__)):
            return getattr(self, 'visit_{0}'.format(type(node).__name__))(node)
        else:
            return self.generic_visit(node)

    def generic_visit(self, node):
        raise Exception("Not Implemented Token: {0}({1!s})".format(type(node).__name__, node))


def call_has_kw_args(node):
    if hasattr(node, "kwargs"):
        return node.kwargs is not None
    return any(kw.arg is None for kw in getattr(node, "keywords", []))


def call_has_starargs(node):
    if hasattr(node, "starargs"):
        return node.starargs is not None
    try:
        # must be defined in Python 3
        ast.Starred
    except AttributeError:
        raise RuntimeError("should not happen")
    return any(isinstance(arg, ast.Starred) for arg in node.args)





# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
Created on Aug 22, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals
import pytest
from hope._transformer import get_fkt_ast
import ast

def dummy(a, b):
    return a + b

@pytest.mark.infrastructure
class TestParser(object):
    
    def test_get_fkt_ast_simple(self):
        fkt_ast = get_fkt_ast(dummy)
        assert fkt_ast is not None
        assert isinstance(fkt_ast, ast.FunctionDef)
        assert len(fkt_ast.body)==1
        assert isinstance(fkt_ast.body[0], ast.Return)
        
    #TODO: how can I cause an exception? Add a new test



# Copyright (c) 2013 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

# setuptools have a proplem with the unicode_literals in python 2, so skip unicode_literals
from __future__ import print_function, division, absolute_import


import os
import io
import sys
import hashlib
import setuptools
import pickle
import inspect
import tempfile
import shutil
import warnings

from numpy.distutils.misc_util import get_numpy_include_dirs
from setuptools.command.build_ext import build_ext
from distutils.dist import Distribution
import distutils.sysconfig


from hope._ast import *
from hope._transformer import ASTTransformer
from hope import config
import hope._cache as cache


class Wrapper:

    _recompile = False

    # the deepcopy of the ast does violates the deepcopy semantics, so do not allow deepcopy ...
    def __deepcopy__(self, memo):
        raise TypeError("The hope function wrapper can't be cloned")

    def __init__(self, fkt, hash):
        self.modtoken, self.fkt, self.filename, self.cache = Module(
            fkt.__name__), fkt, "{0}_{1}".format(fkt.__name__, hash), None

        def recoverArg(arg):
            if isinstance(arg, Object):
                if not hasattr(arg, "parent"):
                    arg.parent = None
                for name, value in list(arg.attrs.items()):
                    if isinstance(value, (Object, ObjectAttr)):
                        value.parent = arg
                    arg.attrs[name] = recoverArg(value)
            else:
                if isinstance(arg.dtype, str) and hasattr(np, arg.dtype):
                    arg.dtype = getattr(np, arg.dtype)
                arg.shape = [(None, Dimension(arg, dim)) for dim in range(arg.dims)]
                delattr(arg, "dims")
                if not isinstance(arg, ObjectAttr):
                    arg.scope, arg.allocated = "signature", True
            return arg

        def create_signature(pickled, args):
            if sys.version_info[0] == 2:
                unpickled = pickle.loads(pickled)
            else:
                import base64
                unpickled = pickle.loads(base64.decodebytes(pickled.encode('ascii')))
            for striped in unpickled:
                signature = [recoverArg(arg) for arg in striped]
                ASTTransformer(self.modtoken).module_visit(self.fkt, signature,
                                                           "".join([arg.getId() for arg in signature]))
            return self(*args)
        self.create_signature = create_signature

        def _hope_callback(*args):
            return self(*args) if self.cache is None else self.cache(*args)
        self.callback = _hope_callback
        setattr(cache, str(id(self.callback)), fkt)

    def __call__(self, *args):
        ASTTransformer(self.modtoken).module_visit(self.fkt, args)

        if config.optimize:

            if config.verbose:
                from hope import _dump
                # TODO: use log
                print(_dump.Dumper().visit(self.modtoken))

            from hope._optimizer import Optimizer
            Optimizer().visit(self.modtoken)

        if config.verbose:
            self._print_module_info()

        try:
            tempfolder = tempfile.mkdtemp(prefix='hope')

            # TODO: if function is class method, make classname_method name as name
            localfilename = "{0}_{1}".format(self.filename, len(
                self.modtoken.functions[self.modtoken.main]) - 1)

            from hope._generator import generate
            code = generate(self.modtoken, localfilename)

            with open(os.path.join(tempfolder, "{0}.cpp".format(localfilename)), "w") as fp:
                fp.write(code)

            so_filename = _compile(tempfolder, localfilename, self.fkt.__name__)

            self._store_state(tempfolder, localfilename)

            # move xxx.so and pickled state if it doesn't already exists or if it was a recompilation due to inconsistent state
            if self._recompile or not os.path.isfile(os.path.join(config.prefix, "{0}.so".format(localfilename))):
                shutil.move(os.path.join(tempfolder, so_filename), os.path.join(
                    config.prefix, "{0}.so".format(localfilename)))
                shutil.move(os.path.join(tempfolder, "{0}.pck".format(self.filename)), os.path.join(
                    config.prefix, "{0}.pck".format(self.filename)))

            if sys.version_info[0] == 2:
                module = __import__(localfilename, globals(), locals(), [], -1)
            else:
                import importlib
                importlib.invalidate_caches()
                module = importlib.import_module(localfilename)
            module.set_create_signature(self.create_signature)

            for name in list(sys.modules.keys()):
                if hasattr(sys.modules[name], self.fkt.__name__) and getattr(sys.modules[name], self.fkt.__name__) is self.callback:
                    setattr(sys.modules[name], self.fkt.__name__, module.run)
            setattr(cache, str(id(module.run)), self.fkt)
            self.cache = module.run

            return module.run(*args)

        finally:
            if config.keeptemp:
                for ext in ["out", "cpp", "o"]:
                    if os.path.isfile(os.path.join(tempfolder, "{0}.{1}".format(localfilename, ext))):
                        shutil.move(os.path.join(tempfolder, "{0}.{1}".format(localfilename, ext)), os.path.join(
                            config.prefix, "{0}.{1}".format(localfilename, ext)))

            shutil.rmtree(tempfolder)

    def _store_state(self, tempfolder, localfilename):
        state = {"filename": localfilename,
                 "main": self.modtoken.main,
                 "called": {},
                 "bound": any([main.isbound for main in self.modtoken.functions[self.modtoken.main]])
                 }

        for name in list(self.modtoken.functions.keys()):
            if name != self.fkt.__name__:
                fkt_hash = None
                # distinguish if a function does not exists or if its an object method
                if not name in self.fkt.__globals__:
                    pass
                #     raise Exception("Function not accessible form global scope of function: {0} ({1})".format(self.fkt.__name__, name))
                # TODO: remove this!
                elif isinstance(self.fkt.__globals__[name], Wrapper):
                    fkt_hash = get_fkt_hash(self.fkt.__globals__[name].fkt)
                elif inspect.isbuiltin(self.fkt.__globals__[name]) and hasattr(cache, str(id(self.fkt.__globals__[name]))):
                    fkt_hash = get_fkt_hash(getattr(cache, str(id(self.fkt.__globals__[name]))))
                elif inspect.isfunction(self.fkt.__globals__[name]):
                    fkt_hash = get_fkt_hash(self.fkt.__globals__[name])
                else:
                    raise Exception("Function not a unbound, pure python function: {0} ({1})".format(
                        self.fkt.__name__, name))

                if fkt_hash is not None:
                    state["called"][name] = fkt_hash

        for name in get_config_attrs():
            state[name] = getattr(config, name)

        with open(os.path.join(tempfolder, "{0}.pck".format(self.filename)), "wb") as fp:
            pickle.dump(state, fp)

    def _print_module_info(self):
        #         from hope import _dump
        # TODO: use log
        #             print(_dump.Dumper().visit(self.modtoken))
        print("Compiling following functions:")
        for (k, v) in self.modtoken.functions.items():
            for node in v:
                signature = []
                for arg in node.signature:
                    if isinstance(arg, Variable):
                        signature.append("{0}{1!s} {2}".format(NPY_TYPE[arg.dtype], "" if len(
                            arg.shape) == 0 else "^{0}".format(len(arg.shape)), arg.name))
                    elif isinstance(arg, Object):
                        signature.append("{0} {1}".format(arg.classname, arg.name))
                    else:
                        raise Exception("Unknown signature argument: {0!s}".format(self.visit(arg)))
                print("{0}({1})".format(node.name, ", ".join(signature)))


def _compile(target, localfilename, fkt_name):
    """
    Compiles a C++ function into a shared object library

    :param target: target folder where the .cpp file is located and where the output should be stored
    :param localfilename: name of file to be compiled without '.cpp' suffix
    :param fkt_name: name of the function to be compiled

    :raises Exception: An exception is raised if the file could be to compiled

    :return so_filename: The name of the .so file
    """
    outfile, stdout, stderr, argv = None, None, None, sys.argv
    is_py_2 = (sys.version_info[0] == 2)
    try:
        sys.stdout.flush(), sys.stderr.flush()
        outfile = open(os.path.join(target, "{0}.out".format(localfilename)), 'w')

        if is_py_2 and  isinstance(sys.stdout, file) and isinstance(sys.stdout, file):
            stdout, stderr = os.dup(sys.stdout.fileno()), os.dup(sys.stderr.fileno())
            os.dup2(outfile.fileno(), sys.stdout.fileno())
            os.dup2(outfile.fileno(), sys.stderr.fileno())
        else:
            stdout, stderr = sys.stdout, sys.stderr
            sys.stdout, sys.stderr = outfile, outfile
        try:
            sources = os.path.join(target, "{0}.cpp".format(localfilename))
            sys.argv = ["", "build_ext",
                        # --build-lib (-b)     directory for compiled extension modules
                        "-b", target,
                        "-t", "/"  # --build-temp - a rel path will result in a dir structure of -b at the cur position
                        ]

            # setuptools have a problem with the unicode_literals in python 2 ...
            if is_py_2:
                from types import StringType
                localfilename = StringType(localfilename)
                sources = StringType(sources)

            # avoid warning on linux + gcc
            cfg_vars = distutils.sysconfig.get_config_vars()
            if "CFLAGS" in cfg_vars:
                cfg_vars["CFLAGS"] = cfg_vars["CFLAGS"].replace("-Wstrict-prototypes", "")
            if "OPT" in cfg_vars:
                cfg_vars["OPT"] = cfg_vars["OPT"].replace("-Wstrict-prototypes", "")

            for p in get_numpy_include_dirs():
                flag = "-I {}".format(p)
                if flag not in config.cxxflags:
                    config.cxxflags.append(flag)
            # config.cxxflags.extend("-I {}".format(p) for p in get_numpy_include_dirs())

            setuptools.setup(
                  name=localfilename,
                  ext_modules=[setuptools.Extension(localfilename,
                                                    sources=[sources],
                                                    extra_compile_args=config.cxxflags)
                               ],
                  include_dirs=get_numpy_include_dirs()
            )
        except SystemExit as e:
            print(sys.stderr.write(str(e)))
        sys.stdout.flush(), sys.stderr.flush()
    finally:
        if isinstance(stdout, int):
            os.dup2(stdout, sys.stdout.fileno()), os.close(stdout)
        elif stdout is not None:
            sys.stdout = stdout
        if isinstance(stderr, int):
            os.dup2(stderr, sys.stderr.fileno()), os.close(stderr)
        elif stderr is not None:
            sys.stderr = stderr
        if (is_py_2 and isinstance(outfile, file))\
                or (not is_py_2 and isinstance(outfile, io.TextIOWrapper) and not outfile.closed):
            outfile.close()
        sys.argv = argv

    with open(os.path.join(target, "{0}.out".format(localfilename))) as outfile:
        out = outfile.read()

    # on Travis CI & Py33 name contains additional suffix to .so
    so_filename = build_ext(Distribution()).get_ext_filename(localfilename)

    if not os.path.isfile(os.path.join(target, so_filename)) or out.find("error:") > -1:
        print(out)
        raise Exception("Error compiling function {0} (compiled to {1})".format(fkt_name, target))

    # TODO: add test case
    if out.find("warning:") > -1:
        try:
            # trying to encode utf-8 to support AstroPy
            warnings.warn("A warning has been issued during compilation:\n{0}".format(
                out).encode('utf-8'))
        except (UnicodeError, TypeError):
            # encoding fails on Linux and Python 3
            warnings.warn("A warning has been issued during compilation:\n{0}".format(out))

    if config.verbose:
        print(out)

    return so_filename


def get_config_attrs():
    """
    Returns the attributes of the hope config, filtering private attributes and imports from __future__

    :return generator: a generator yielding the attributes
    """
    return (name for name in dir(config) if (not name.startswith("__") and name not in ("print_function", "division", "absolute_import", "unicode_literals")))


def get_fkt_hash(fkt):
    """
    Returns hash of the `uft-8` encoded source of the given function

    :param fkt: the function to compute the hash from

    :return hash: the sha224 hash code 
    """
    return hashlib.sha224(inspect.getsource(fkt).encode('utf-8')).hexdigest()



# Copyright (c) 2013 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals

import subprocess
from distutils.version import StrictVersion
from distutils.sysconfig import customize_compiler
from distutils import sysconfig

from hope.exceptions import UnsupportedCompilerException

CXX_FLAGS = {
    "clang": ["-Wall", "-Wno-unused-variable", "-march=native", "-stdlib=libc++", "-std=c++11",
              "-Wno-unreachable-code"],
    "icc": ["-Wall", "-Wno-unused-variable", "-march=native", "-stdlib=libc++", "-std=c++11",
            "-Wno-unreachable-code"],
    "gcc-mac": ["-Wall", "-Wno-unused-variable", "-std=c++11", "-msse4.2", "-Wno-unreachable-code"],
    "gcc-linux": ["-Wall", "-Wno-unused-variable", "-std=c++11", "-Wno-unreachable-code"]
}

DARWIN_KEY = "Darwin"
LINUX_KEY = "Linux"

GCC_CLANG_VERSION = "4.2.1"  # always returned by clang -dumpversion
MIN_GCC_VERSION = "4.7.0"
SUPPORTED_VERSIONS = {"gcc": MIN_GCC_VERSION,
                      "gcc-linux": MIN_GCC_VERSION,
                     }


def enableUnsaveMath():
    """
    Enable the fast-math and associative-math flags in the c++ compiler.

    .. warning::
        using these flags, ``nan`` and ``inf`` are not propageted properly.
        Only use these flags if you know what you are doing!
    """
    from . import config
    config.cxxflags += ["-fassociative-math", "-ffast-math"]


# Disable faster but unsecure Math
def disableUnsaveMath():
    """
    Disable the fast-math and associative-math flags in the c++ compiler.
    """
    from . import config

    flags = config.cxxflags
    for flag in ("-ffast-math", "-fassociative-math"):
        if flag in flags:
            flags.remove(flag)


# HOPE requires the compiler to compile with c++11 features enabled
def get_cxxflags():

    # late imports here make mocking in tests possible:
    from platform import system
    from distutils.ccompiler import new_compiler

    if system() == DARWIN_KEY:
        CXX_FLAGS["gcc"] = CXX_FLAGS["gcc-mac"]
        CXX_FLAGS["cc"] = CXX_FLAGS["clang"]
        CXX_FLAGS["c++"] = CXX_FLAGS["clang"]
    elif system() == LINUX_KEY:
        CXX_FLAGS["gcc"] = CXX_FLAGS["gcc-linux"]
        CXX_FLAGS["cc"] = CXX_FLAGS["gcc"]
        CXX_FLAGS["c++"] = CXX_FLAGS["gcc"]
    else:
        raise UnsupportedCompilerException("System: %s is not supported by HOPE" % system())

    sysconfig.get_config_vars()  # init vars
    compiler = new_compiler()
    customize_compiler(compiler)
    compiler_name = compiler.compiler[0].split("/")[-1]

    if compiler_name not in CXX_FLAGS.keys():  # trying to support x86_64-linux-gnu-gcc
        compiler_name = "gcc-linux" if compiler_name.find("gcc") > -1 else compiler_name

    _check_version(compiler_name, compiler.compiler[0])

    for name, flags in CXX_FLAGS.items():
        if compiler_name.startswith(name):
            return flags
    raise UnsupportedCompilerException("Unknown compiler: {0}".format(compiler_name))


def _check_version(compiler_name, compiler_exec):
    if compiler_name in SUPPORTED_VERSIONS.keys():
        flags = ('-dumpversion', '-dumpfullversion')
        for flag in flags:
            try:
                sversion = subprocess.check_output(
                    compiler_exec + ' ' + flag, shell=True).decode().rstrip()
                version = str(StrictVersion(sversion))
                break
            except ValueError:
                continue
        else:
            raise ValueError(
                'gcc did not provide usefull version info for any of the flags {}'.format(flags))

        if version < StrictVersion(SUPPORTED_VERSIONS[compiler_name]):
            # dont raise an exception if its gcc proxied clang
            if version != StrictVersion(GCC_CLANG_VERSION):
                msg = ("Compiler '%s' with version '%s' is not supported. Minimum version is '%s'"
                       % (compiler_exec, sversion, SUPPORTED_VERSIONS[compiler_name]))
                raise UnsupportedCompilerException(msg)



# Copyright (C) 2013 ETH Zurich, Institute for Astronomy

'''
Created on Aug 8, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals
from hope import _wrapper
import os
import pytest
from py._path.local import LocalPath
from hope._ast import FunctionDef

def dummy_fkt():
    return 0

@pytest.mark.infrastructure
class TestWrapper(object):
    
    cpp_hello_world = """
                        int main()
                        {
                            return 0;
                        }"""

    def test_compile(self, tmpdir):
        tmp_path = str(tmpdir)
        localfilename = "helloworld"
        with open(os.path.join(tmp_path, localfilename+".cpp"), "w") as source:
            source.write(self.cpp_hello_world)
        
        _wrapper._compile(tmp_path, localfilename, localfilename)
        
        assert os.path.exists(os.path.join(tmp_path, localfilename+".out"))
        
        
    def test_compile_invalid(self, tmpdir):
        tmp_path = str(tmpdir)
        localfilename = "invalid"
        with open(os.path.join(tmp_path, localfilename+".cpp"), "w") as source:
            source.write("invalid")
        
        try:
            _wrapper._compile(tmp_path, localfilename, localfilename)
            pytest.fail("Source can't be compiled!")
        except Exception:
            assert True        


    def test_store_state(self, tmpdir):
        tmp_path = str(tmpdir)
        localfilename = "dummy_fkt"
        
        wrapper = _wrapper.Wrapper(dummy_fkt, "0")
        wrapper.modtoken.functions[localfilename] = [FunctionDef(localfilename, [])]
        wrapper._store_state(tmp_path, localfilename)
        
        assert os.path.exists(os.path.join(tmp_path, "{0}.pck".format(wrapper.filename)))
        



# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals


import sys
import pickle

from hope import config
from hope._ast import *
from hope._const import *
from hope._library import *
from hope._dump import Dumper

class CPPGenerator(NodeVisitor):
    """
    Generates the C code from the given :py:class:`hope._ast.Module` token
    by traversing the AST.
    """

    def __init__(self):
        self.next_loopid, self.merged, self.slicemap, self.library, self.dumper  = 0, None, {}, {}, Dumper()

    def getVariableExtent(self, node):
        extent = ""
        for ind, segment in enumerate(node.shape):
            segmentstr = self.get_segmentstr(*segment)
            if not segment[0] is None:
                raise Exception("Variable slices needs to start with None: {0}".format(node.name))
            if not segmentstr in self.merged:
                raise Exception("Unknown slice {0} in variable {1}".format(segmentstr, node.name))
            if ind > 0:
                extent = "({0})*{1}".format(extent, self.visit(segment[1]))
            extent += "{0}{1}".format(" + " if len(extent) > 0 else "", self.slicemap[self.get_slicemap_key(ind, *segment)])
        return extent

    def visit_Number(self, node):
        if node.dtype is bool:
            return "true" if node.value else "false"
        if config._readablecxx:
            return "{1!r}".format(PY_C_TYPE[node.dtype], node.value)
        else:
            return "({0}){1!r}".format(PY_C_TYPE[node.dtype], node.value)

    def visit_Variable(self, node):
        if len(node.shape) == 0 or node.scope == "block":
            return "c{0}".format(node.name)
        else:
            return "c{0}[{1}]".format(node.name, self.getVariableExtent(node));

    def visit_ObjectAttr(self, node):
        trace = node.getTrace()
        if len(node.shape) == 0:
            return "c" + ".c".join(trace)
        else:
            return "{0}[{1}]".format("c" + ".c".join(trace), self.getVariableExtent(node));

    def visit_Dimension(self, node):
        if isinstance(node.variable, ObjectAttr):
            parent = node.variable.parent
            trace = ["s{0}".format(node.variable.name)]
            while not parent is None:
                trace.insert(0, "c{0}".format(parent.name))
                parent = parent.parent
            return "{0}[{1}]".format(".".join(trace), node.dim)
        elif isinstance(node.variable, Variable):
            return "s{0}[{1}]".format(node.variable.name, node.dim)
        else:
            raise Exception("Unknown type {0}".format(node.variable.name))

    def visit_DimensionSlice(self, node):
        return "{0}+{1}".format(self.visit_Dimension(node), self.visit(node.slice))

    def visit_View(self, node):
        subscript = ""
        extent_ind = 0
        for ind, (extent, segment) in enumerate(zip(node.extents, node.variable.shape)):
            if ind > 0:
                subscript = "(int)({0})*{1}".format(subscript, self.visit(segment[1]))
            subscript += " + " if len(subscript) > 0 else ""
            if isinstance(extent, tuple):
                lower, upper = extent
                if lower is None: lower = segment[0]
                if isinstance(lower, Number) and lower.value == 0: lower = None
                if upper is None: upper = segment[1]
                seg = "{0} + ".format(self.visit(lower)) if not lower is None else ""
                key = self.get_slicemap_key(extent_ind, lower, upper)
                seg += self.slicemap[key]
#                 seg += self.slicemap[self.merged["{0}:{1}".format("" if lower is None else self.dumper.visit(lower), self.dumper.visit(upper))]]
                segstr = "{0}:{1}".format("" if lower is None else self.dumper.visit(lower), self.dumper.visit(upper))
                extent_ind += 1
            else:
                seg = self.visit(extent)
                if isinstance(extent, Number) and extent.value < 0:
                    seg = "{0}+{1}".format(self.visit_Dimension(segment[1]), seg)
                segstr = self.dumper.visit(extent)
            if config.rangecheck:
                subscript += "native_rangecheck({0}".format(seg)
                subscript += ", {0}, {1}".format(self.visit(segment[0]) if not segment[0] is None else "0", self.visit(segment[1]))
                subscript += ", std::string(\"{0}\"), std::string(\"{1}\"))".format(segstr, node.variable.name)
                self.library["native_rangecheck"] = LIBRARY_NATIVE_RANGECHECK
            else:
                subscript += seg
        shape = node.variable.shape
        node.variable.shape = []
        # TODO: if double, cast to int ...
        ret = "{0}[(int)({1})]".format(self.visit(node.variable), subscript);
        node.variable.shape = shape
        return ret

    def visit_Expr(self, node):
        return "{0};".format(self.visit(node.value))

    def visit_NumpyContraction(self, node):
        if node.op != "sum":
            raise Exception("Only the numpy.sum contraction is implemented!")
        if len(node.value.shape):
            ret = "{0} = ({1})0;\n".format(self.visit(node.variable), PY_C_TYPE[node.dtype])
            keys = []
            for ind, segment in enumerate(node.value.shape):
                ret += "{0}for (npy_intp i{1} = 0; i{1} < {2} - {3}; ++i{1}) {{\n".format( \
                      "\t" * ind \
                    , self.next_loopid \
                    , self.visit(segment[1]) \
                    , 0 if segment[0] is None else self.visit(segment[0]) \
                )
                segmentstr = self.get_segmentstr(*segment)
                if not segmentstr in self.merged:
                    self.merged[segmentstr] = segmentstr
                keys.append(self.get_slicemap_key(ind, *segment))
                self.slicemap[keys[-1]] = "i{0}".format(self.next_loopid)
                self.next_loopid += 1
            ret += "{0}{1} += {2};".format("\t" * len(node.value.shape), self.visit(node.variable), self.visit(node.value))
            for ind, (key, segment) in enumerate(zip(keys, node.value.shape)):
                del self.slicemap[key]
                ret += "\n{0}}}".format("\t" * (len(node.value.shape) - 1 - ind))
            return ret
        else:
            return "\n{0} = {1};".format(self.visit(node.variable), self.visit(node.value))

    def visit_Assign(self, node):
        # TODO: capture type
        if isinstance(node.target, Variable) and node.target.allocated == False:
            node.target.allocated = True
            return "auto {0} = {1};".format(self.visit(node.target), self.visit(node.value))
        else:
            return "{0} = {1};".format(self.visit(node.target), self.visit(node.value))

    def visit_Reference(self, node):
        target = node.target
        trace = node.value.getTrace()
        if isinstance(target, ObjectAttr): # self.x = self.y
            return "c{0} = c{1};".format(".c".join(target.getTrace()), ".c".join(trace))
        if len(target.shape) == 0: # [int] x = self.y
            return "{0} c{1} = c{2};".format(PY_C_TYPE[target.dtype], target.name, ".c".join(trace))
        else: # [array] x = self.y
            return "PyObject * p{0} = (PyObject *)PyArray_GETCONTIGUOUS((PyArrayObject *)c{1});\n".format(target.name, ".p".join(trace)) \
                +  "npy_intp * s{0} = c{1};\n".format(target.name, ".s".join(trace)) \
                +  "{0} * c{1} = c{2};".format(PY_C_TYPE[target.dtype], target.name, ".c".join(trace))

    def visit_AugAssign(self, node):
        if node.op == "**=":
            return "{0} = std::pow({0}, {1});".format(self.visit(node.target), self.visit(node.value))
        elif node.op == "//=":
            if type(node.target.dtype(1) // node.value.dtype(1)) in [float, np.float32, np.float64]:
                return "{0} = std::floor({0} / {1});".format(self.visit(node.target), self.visit(node.value))
            else:
                return "{0} /= {1};".format(self.visit(node.target), self.visit(node.value))
        elif node.op == "%=":
            self.library["native_mod"] = LIBRARY_NATIVE_MOD
            return "{0} = native_mod({0}, {1});".format(self.visit(node.target), self.visit(node.value))
        else:
            return "{0} {1} {2};".format(self.visit(node.target), node.op, self.visit(node.value))

    def visit_UnaryOp(self, node):
        if config._readablecxx:
            return "{0}{1}".format(node.op, self.visit(node.operand))
        else:
            return "({0}{1})".format(node.op, self.visit(node.operand))

    def visit_BinOp(self, node):
        cast = "" if node.dtype is None or config._readablecxx else "({0})".format(PY_C_TYPE[node.dtype])
        if node.op == "**":
            return "{0}std::pow({1}, {2})".format(cast, self.visit(node.left), self.visit(node.right))
        elif node.op == "//" and node.dtype in [float, np.float32, np.float64]:
            return "{0}std::floor({1} / {2})".format(cast, self.visit(node.left), self.visit(node.right))
        elif node.op == "//":
            return "{0}({1} / {2})".format(cast, self.visit(node.left), self.visit(node.right))
        else:
            left = self.visit(node.left) if node.dtype == node.left else "{0}{1}".format(cast, self.visit(node.left))
            right = self.visit(node.right) if node.dtype == node.right else "{0}{1}".format(cast, self.visit(node.right))
            if node.op == "%":
                self.library["native_mod"] = LIBRARY_NATIVE_MOD
                return "native_mod({0}, {1})".format(left, right)
            else:
                return "{0}({1} {2} {3})".format(cast, left, node.op, right)

    def visit_BoolOp(self, node):
        values = [self.visit(value) for value in node.values]
        return "({0})".format(" {0} ".format(node.op).join(values))

    def visit_Compare(self, node):
        cast = "" if node.dtype is None or config._readablecxx else "({0})".format(PY_C_TYPE[node.dtype])
        return "{0}({1} {2} {3})".format(cast, self.visit(node.left), node.op, self.visit(node.comparator))

    def visit_If(self, node):
        ret = "if ({0}) {{\n\t{1}\n}}".format(self.visit(node.test), "\n\t".join(self.visit(node.body).split("\n")))
        if not node.orelse is None:
            ret += " else {{\n\t{0}\n}}".format("\n\t".join(self.visit(node.orelse).split("\n")))
        return ret

    def visit_For(self, node):
        ret = "for (npy_intp {0} = {1}; {0} < {2}; ++{0}) {{\n\t".format(self.visit(node.iter), self.visit(node.lower), self.visit(node.upper))
        ret += "\n\t".join(self.visit(node.body).split("\n"))
        ret += "\n}"
        return ret

    def visit_While(self, node):
        ret = "while ({0}) {{\n\t".format(self.visit(node.test))
        ret += "\n\t".join(self.visit(node.body).split("\n"))
        ret += "\n}"
        return ret

    def visit_Call(self, node):
        # TODO: generalize
        if isinstance(node.name, GlobalFunction):
            args = []
            for arg in node.args:
                if not isinstance(arg, Object) and len(arg.shape) > 0 and (not isinstance(arg, Variable) or arg.scope == "block"):
                    raise Exception("Only variables can be passed to funtions!")
                elif isinstance(arg, Object):
                    args.append("c{0}".format(arg.name))
                elif len(arg.shape) > 0:
                    args.append("p{0}, s{0}, c{0}".format(arg.name))
                else:
                    args.append(self.visit(arg))
            return "{0}_{1}({2})".format( \
                  node.name.name \
                , "".join([arg.getId() for arg in node.args]) \
                , ", ".join(args) \
            )
        elif isinstance(node.name, HopeAttr) and node.name.name == "exp":
            self.library["hope_exp"] = LIBRARY_HOPE_EXP
            return "hope_exp({0})".format(self.visit(node.args[0]))
        elif isinstance(node.name, NumpyAttr) and node.name.name in ["empty", "zeros", "ones"]:
            return "1" if node.name.name == "ones" else "0"
        elif isinstance(node.name, NumpyAttr) and node.name.name == "interp":
            self.library["numpy_interp"] = LIBRARY_NUMPY_INTERP

            args = [[] for _ in range(3)]
            for i in range(1, 3):
                if isinstance(node.args[i], ObjectAttr):
                    parent, trace = node.args[i].parent, [node.args[i].name]
                    while not parent is None:
                        trace.insert(0, parent.name)
                        parent = parent.parent
                    args[i] = ".c".join(trace)
                else:
                    args[i] = node.args[i].name
                    
            # TODO: make sure node.args[1].shape == node.args[2].shape using an assert
            
            ((lower, upper),) = node.args[1].shape
            left_val = "c{0}[0]".format(node.args[1].name) if lower is None else self.visit(lower)
            size = self.visit(upper) + ("" if lower is None else "-{0}".format(self.visit(lower)))
            right_val = "c{0}[{1}]".format(node.args[1].name, size + "-1")

            ret = "numpy_interp({0}, c{1}, c{2}, {3})".format(self.visit(node.args[0]), args[1], args[2], size)
            
            if "left" in node.keywords:
                left_ret = self.visit(node.keywords["left"])
            else:
                left_ret = "c{0}[0]".format(args[2])
                
            ret = "{0} < {1} ? {2} : ({3})".format(self.visit(node.args[0]), left_val, left_ret, ret)
                
            if "right" in node.keywords:
                right_ret = self.visit(node.keywords["right"])
            else:
                right_ret = "c{0}[{1}]".format(args[2], size + "-1")
                
            ret = "{0} > {1} ? {2} : ({3})".format(self.visit(node.args[0]), right_val, right_ret, ret)
                
            return "({0})".format(ret)
        elif isinstance(node.name, NumpyAttr) and node.name.name == "sign":
            self.library["native_sign"] = LIBRARY_NATIVE_SIGN
            return "native_sign({0})".format(self.visit(node.args[0]))
        elif isinstance(node.name, NumpyAttr) and node.name.name in NPY_UNARY_FUNCTIONS:
            return "{0}({1})".format(NPY_UNARY_FUNCTIONS[node.name.name], self.visit(node.args[0]))
        elif isinstance(node.name, NumpyAttr) and node.name.name in NPY_CAST_FUNCTIONS:
            return "({0})({1})".format(PY_C_TYPE[NPY_CAST_FUNCTIONS[node.name.name]], self.visit(node.args[0]))

    def visit_Allocate(self, node):
        shape, variable = [], node.variable
        if variable.dtype is None:
            raise Exception("Unknown dtype: {0}".format(variable.dtype))
        for segment in variable.shape:
            if not segment[0] is None:
                raise Exception("Allocate need to have (:len)* in shape: {0}".format(",".join([str(sgment) for sgment in variable.shape])))
            shape.append(self.visit(segment[1]))
        if len(shape) == 0:
            return "{0} c{1} = {0}();".format(PY_C_TYPE[variable.dtype], variable.name)
        else:
            return "npy_intp d{0}[] = {{(npy_intp){1}}};\n".format(variable.name, ", (npy_intp)".join(shape)) \
                +  "PyObject * p{0} = PyArray_EMPTY({1}, d{0}, {2}, 0);\n".format(variable.name, len(shape), NPY_TYPEENUM[variable.dtype]) \
                +  "npy_intp * s{0} = PyArray_SHAPE((PyArrayObject *)p{0});\n".format(variable.name) \
                +  "{0} * c{1} = ({0} *)PyArray_DATA((PyArrayObject *)p{1});".format(PY_C_TYPE[variable.dtype], variable.name)

    def visit_Return(self, node):
        # TODO: implement expressions
        if len(node.value.shape) == 0:
            return "return {0};".format(self.visit(node.value));
        elif not isinstance(node.value, Variable):
            raise Exception("TODO: implement!")
        else:
            return "return std::make_tuple((PyObject *)p{0}, s{0}, c{0});".format(node.value.name);

    def visit_Block(self, node):
        if len(node.shape):
            ret = "";
            keys = []
            for ind, segment in enumerate(node.shape):
                ret += "{0}for (npy_intp i{1} = 0; i{1} < {2} - {3}; ++i{1}) {{\n".format(
                          "\t" * ind, 
                          self.next_loopid, 
                          self.visit(segment[1]), 
                          0 if segment[0] is None else self.visit(segment[0])
                        )
                keys.append(self.get_slicemap_key(ind, *segment))
                self.slicemap[keys[-1]] = "i{0}".format(self.next_loopid)
                self.next_loopid += 1
                
            ret += "{0}".format("\t" * len(node.shape))
            ret += "\n{0}".format("\t" * len(node.shape)).join("\n".join([self.visit(expr) for expr in node.body]).split("\n"))
            
            for ind, key in enumerate(keys):
                del self.slicemap[key]
                ret += "\n{0}}}".format("\t" * (len(node.shape) - 1 - ind))
                
            return ret
        else:
            return "\n".join(["{0}".format(self.visit(expr)) for expr in node.body])

    def visit_Body(self, node):
        return "\n".join([self.visit(block) for block in node.blocks])

    def visit_FunctionDef(self, node):
        self.merged, code, firstSegment = {}, "", None
        for merged in node.merged:
            if len(merged) == 1:
                self.merged[merged[0]] = merged[0]
            else:
                for segment in merged:
                    self.merged[segment] = merged[0]
                    if isinstance(node.shapes[segment], tuple):
                        lower, upper = node.shapes[segment]
                        if isinstance(lower, Number) and lower.value == 0: lower = None
                        if upper is None:
                            raise Exception("Unbound shapes cannot be merged: {0}".format(segment))
                        if lower is None:
                            shape = self.visit(upper)
                        else:
                            shape = "({1} - {0})".format(self.visit(lower), self.visit(upper))
                    else:
                        raise Exception("Indexes can not be merged: {0!s}".format(segment))
                    if firstSegment is None:
                        firstSegment = shape
                    # TODO: this fails if variable is defined in loop, maybe it makes sense to visit lower and upper und check if we can check ...
                    elif config.rangecheck:
                        code += "\n\tif ({0} - {1} != 0) {{".format(firstSegment, shape)
                        code += "\n\t\tPyErr_SetString(PyExc_ValueError, \"Shapes {0} and {1} do not match!\");".format(firstSegment, shape)
                        code += "\n\t\tthrow std::exception();"
                        code += "\n\t}"
        code += "\n\t" + "\n\t".join(self.visit(node.body).split("\n"))
        if not node.dtype is None:
            code += "\n\tPyErr_SetString(PyExc_ValueError, \"No return type passed!\");"
            code += "\n\tthrow std::exception();"
        return code

    def visit_Module(self, node):
        for fktname, fktlist in list(node.functions.items()):
            for fkt in fktlist:
                fkt.decl, sig = "inline ", []
                if fkt.dtype is None:
                    fkt.decl += "void"
                elif len(fkt.shape) == 0:
                    fkt.decl += PY_C_TYPE[fkt.dtype]
                else:
                    fkt.decl += "std::tuple<PyObject *, npy_intp const *, {0} *>".format(PY_C_TYPE[fkt.dtype])
                fkt.decl += " {0}_{1}(\n\t  ".format(fktname, fkt.getId())
                for arg in fkt.signature:
                    if isinstance(arg, Object):
                        sig.append("{0} & c{1}\n".format(arg.getId("t"), arg.name))
                    elif len(arg.shape) > 0:
                        sig.append("PyObject * p{1}, npy_intp const * __restrict__ s{1}, {0} * __restrict__ c{1}\n".format(PY_C_TYPE[arg.dtype], arg.name))
                    else:
                        sig.append("{0} c{1}\n".format(PY_C_TYPE[arg.dtype], arg.name))
                fkt.decl += "\t, ".join(sig) + ")"
        code = "".join([fkt.decl + ";\n" for fktname, fktlist in list(node.functions.items()) for fkt in fktlist])

        for fktname, fktlist in list(node.functions.items()):
            for fkt in fktlist:
                code += fkt.decl + " {"
                code += self.visit(fkt)
                code += "\n}\n"
        return "\n".join(list(self.library.values())) + code


    def get_segmentstr(self, lower, upper):
        return "{0}:{1}".format("" if lower is None else self.dumper.visit(lower), self.dumper.visit(upper))
    
    def get_slicemap_key(self, ind, lower, upper):
        segmentstr = self.get_segmentstr(lower, upper)
        return "i{0}>{1}".format(ind, self.merged[segmentstr])

def generate(modtoken, localfilename):
    """
    Generates the C code from the given :py:class:`hope._ast.Module` token
    
    :param modtoken: Module to use
    :param localfilename: name of the function incl. signature
    
    :return code: the generated C code
    """

    objects = []
    def findObjects(obj):
        if obj.getId() not in [arg.getId() for arg in objects]:
            objects.insert(0, obj)
        for variable in list(obj.attrs.values()):
            if isinstance(variable, Object):
                findObjects(variable)

    for fktlist in list(modtoken.functions.values()):
        for fkt in fktlist:
            for arg in fkt.signature:
                if isinstance(arg, Object):
                    findObjects(arg)

    code  = LIBRARY_IMPORTS
    code += LIBRARY_PYOBJ_DEF
    
    code += _obj_init_code(objects)
    
    generator = CPPGenerator()
    code += generator.visit(copy.deepcopy(modtoken))

    code += LIBRARY_SIGHANDLER

    code += "\n"
    code += "extern \"C\" {\n"
    
    code += LIBRARY_CREATE_SIGNATURE

    code += _run_fkt_code(modtoken)
    
    code += "\t}\n"
    #end of extern block

    if sys.version_info[0] == 2:
        code += LIBRARY_METHODS_DECL_PY2.format(fktname=modtoken.main)
        code += LIBRARY_INIT_DECL_PY2.format(filename=localfilename, fktname=modtoken.main)
        
    else:
        code += LIBRARY_METHODS_DECL_PY3.format(fktname=modtoken.main)
        code += LIBRARY_MODULE_DECL_PY3.format(fktname=modtoken.main)
        code += LIBRARY_INIT_DECL_PY3.format(filename=localfilename, fktname=modtoken.main)

    code += "}\n"

    return code

def _obj_init_code(objects):
    code = ""
    for obj in objects:

        code += "struct {0} {{\n".format(obj.getId("t"))

        code += "\tbool initialize(PyObject * obj) {"
        if len(obj.attrs) == 0:
            code += "\n\t\treturn true;"
        else:
            code += "\n\t\tif ("
            for pos, (name, variable) in enumerate(obj.attrs.items()):
                code += "\n\t\t\t"
                if pos > 0:
                    code += "and "
                code += "PyObject_HasAttrString(obj, \"{0}\") and p{0}.incref(PyObject_GetAttrString(obj, \"{0}\")) ".format(name)
                if isinstance(variable, Object):
                    code += "and c{0}.initialize(p{0})".format(name)
                elif len(variable.shape) > 0:
                    code += "and PyArray_CheckExact(p{0})".format(name)
                    code += "\n\t\t\tand PyArray_TYPE((PyArrayObject *)p{0}) == {1} and PyArray_NDIM((PyArrayObject *)p{0}) == {2}".format(name, NPY_TYPEENUM[variable.dtype], len(variable.shape))
                elif variable.dtype is int:
                    if sys.version_info[0] == 2:
                        code += "and PyInt_CheckExact((PyObject *)p{0})".format(name)
                    else:
                        code += "and PyLong_CheckExact((PyObject *)p{0})".format(name)
                elif variable.dtype is float:
                    code += "and PyFloat_CheckExact((PyObject *)p{0})".format(name)
                elif variable.dtype in NPY_SCALAR_TAG:
                    code += "and PyArray_IsScalar((PyArrayObject *)p{0}, {1})".format(name, NPY_SCALAR_TAG[variable.dtype])
                else:
                    raise Exception("Unknown type: {0!s}".format(variable.dtype))
            code += "\n\t\t) {\n"
            for name, variable in list(obj.attrs.items()):
                if isinstance(variable, Object):
                    pass
                elif len(variable.shape) > 0:
                    code += "\t\t\tif (!(p{0}.incref((PyObject *)PyArray_GETCONTIGUOUS((PyArrayObject *)p{0})))) {{\n".format(name)
                    code += "\t\t\t\tPyErr_SetString(PyExc_ValueError, \"Invalid Argument type on {0}!\");\n".format(name)
                    code += "\t\t\t\tthrow std::exception();\n"
                    code += "\t\t\t}\n"
                    code += "\t\t\ts{0} = PyArray_SHAPE((PyArrayObject *)p{0});\n".format(name)
                    code += "\t\t\tc{1} = ({0} *)PyArray_DATA((PyArrayObject *)p{1});\n".format(PY_C_TYPE[variable.dtype], name)
                elif variable.dtype is int:
                    if sys.version_info[0] == 2:
                        code += "\t\t\tc{0} = PyInt_AS_LONG((PyObject *)p{0});\n".format(name)
                    else:
                        code += "\t\t\tc{0} = PyLong_AS_LONG((PyObject *)p{0});\n".format(name)
                elif variable.dtype is float:
                    code += "\t\t\tc{0} = PyFloat_AS_DOUBLE((PyObject *)p{0});\n".format(name)
                elif variable.dtype in NPY_SCALAR_TAG:
                    code += "\t\t\tc{0} = PyArrayScalar_VAL((PyArrayObject *)p{0}, {1});\n".format(name, NPY_SCALAR_TAG[variable.dtype])
            code += "\t\t\treturn true;\n"
            code += "\t\t} else\n"
            code += "\t\t\treturn false;\n"

        code += "\t}\n"
        for name, variable in list(obj.attrs.items()):
            if isinstance(variable, Object):
                code += "\tPyObj p{1};\n\t{0} c{1};\n".format(variable.getId("t"), name)
            elif len(variable.shape) > 0:
                code += "\tPyObj p{1};\n\tnpy_intp * s{1};\n\t{0} * __restrict__ c{1};\n".format(PY_C_TYPE[variable.dtype], name)
            else:
                code += "\tPyObj p{1};\n\t{0} c{1};\n".format(PY_C_TYPE[variable.dtype], name)
        code += "};\n"

    return code

def _run_fkt_code(modtoken):
    code = ""
    code += "\tPyObject * run(PyObject * self, PyObject * args) {\n"

    for fkt in modtoken.functions[modtoken.main]:
        code += "\t\t{"
        for arg in fkt.signature:
            if isinstance(arg, Object):
                code += "\n\t\t\tPyObject * p{0};".format(arg.name)
                code += "\n\t\t\t{0} c{1};".format(arg.getId("t"), arg.name)
            elif len(arg.shape) > 0:
                code += "\n\t\t\tPyObj p{0};".format(arg.name)
            else:
                code += "\n\t\t\tPyObject * p{0};".format(arg.name)
                code += " {0} c{1};".format(PY_C_TYPE[arg.dtype], arg.name)
        if len(fkt.signature) > 0:
            code += "\n\t\t\tif ("
            code += "\n\t\t\t\tPyTuple_CheckExact(args) and PyTuple_GET_SIZE(args) == {0}".format(len(fkt.signature))
            for idx, arg in enumerate(fkt.signature):
                code += "\n\t\t\t\tand (p{0} = PyTuple_GET_ITEM(args, {1})) ".format(arg.name, idx)
                if isinstance(arg, Object):
                    code += "and c{0}.initialize(p{0})".format(arg.name)
                elif len(arg.shape) > 0:
                    code += "and PyArray_CheckExact(p{0})".format(arg.name)
                    code += "\n\t\t\t\tand PyArray_TYPE((PyArrayObject *)p{0}) == {1} and PyArray_NDIM((PyArrayObject *)p{0}) == {2}".format(arg.name, NPY_TYPEENUM[arg.dtype], len(arg.shape))
                elif arg.dtype is int:
                    if sys.version_info[0] == 2:
                        code += "and PyInt_CheckExact(p{0})".format(arg.name)
                    else:
                        code += "and PyLong_CheckExact(p{0})".format(arg.name)
                elif arg.dtype is float:
                    code += "and PyFloat_CheckExact(p{0})".format(arg.name)
                elif arg.dtype in NPY_SCALAR_TAG:
                    code += "and PyArray_IsScalar(p{0}, {1})".format(arg.name, NPY_SCALAR_TAG[arg.dtype])
                else:
                    raise Exception("Unknown type: {0!s}".format(arg.dtype))

            code += "\n\t\t\t) {\n"
            for arg in fkt.signature:
                if isinstance(arg, Object): pass
                elif len(arg.shape) > 0:
                    code += "\t\t\t\tif (!(p{0}.incref((PyObject *)PyArray_GETCONTIGUOUS((PyArrayObject *)p{0})))) {{\n".format(arg.name)
                    code += "\t\t\t\t\tPyErr_SetString(PyExc_ValueError, \"Invalid Argument type on {0}!\");\n".format(arg.name)
                    code += "\t\t\t\t\treturn NULL;\n"
                    code += "\t\t\t\t}\n"
                elif arg.dtype is int:
                    if sys.version_info[0] == 2:
                        code += "\t\t\t\tc{0} = PyInt_AS_LONG(p{0});\n".format(arg.name)
                    else:
                        code += "\t\t\t\tc{0} = PyLong_AS_LONG(p{0});\n".format(arg.name)
                elif arg.dtype is float:
                    code += "\t\t\t\tc{0} = PyFloat_AS_DOUBLE(p{0});\n".format(arg.name)
                elif arg.dtype in NPY_SCALAR_TAG:
                    code += "\t\t\t\tc{0} = PyArrayScalar_VAL(p{0}, {1});\n".format(arg.name, NPY_SCALAR_TAG[arg.dtype])

            args = []
            for arg in fkt.signature:
                if not isinstance(arg, Object) and len(arg.shape) > 0:
                    args.append("p{1}, PyArray_SHAPE((PyArrayObject *)p{1}), ({0} *)PyArray_DATA((PyArrayObject *)p{1})".format(PY_C_TYPE[arg.dtype], arg.name))
                else:
                    args.append("c{0}".format(arg.name))

            call  = "{0}_{1}(".format(modtoken.main, fkt.getId())
            call += "\n\t\t\t\t\t\t  {0}".format("\n\t\t\t\t\t\t, ".join(args))
            call += "\n\t\t\t\t\t)"
        else:
            call  = "{0}_{1}()".format(modtoken.main, fkt.getId())

        code += "\t\t\t\ttry {\n"
        if fkt.dtype is None:
            code += "\t\t\t\t\t{0};\n".format(call)
            code += "\t\t\t\t\tPy_INCREF(Py_None);\n"
            code += "\t\t\t\t\treturn Py_None;\n"
        elif len(fkt.shape) == 0 and fkt.dtype is bool:
            code += "\t\t\t\t\tPyObject* res = {0} ? Py_True : Py_False;\n".format(call)
            code += "\t\t\t\t\tPy_INCREF(res);\n"
            code += "\t\t\t\t\treturn res;\n"
        elif len(fkt.shape) == 0 and fkt.dtype is int:
            code += "\t\t\t\t\treturn Py_BuildValue(\"{0}\", {1});\n".format(PY_TYPE_CHAR[np.int_], call)
        elif len(fkt.shape) == 0 and fkt.dtype is float:
            code += "\t\t\t\t\treturn Py_BuildValue(\"{0}\", {1});\n".format(PY_TYPE_CHAR[np.float_], call)
        elif len(fkt.shape) == 0 and fkt.dtype in NPY_SCALAR_TAG:
            code += "\t\t\t\t\tPyObject* res = PyArrayScalar_New({0});\n".format(NPY_SCALAR_TAG[fkt.dtype])
            code += "\t\t\t\t\tPyArrayScalar_ASSIGN(res, {0}, {1});\n".format(NPY_SCALAR_TAG[fkt.dtype], call)
            code += "\t\t\t\t\treturn res;\n"
        else:
            code += "\t\t\t\t\tPyObject * res = std::get<0>({0});\n".format(call)
            if fkt.return_allocated:
                # to avoid mem leak or segfault
                code += "\n\t\t\t\t\tPy_INCREF(res);\n"
                
            code += "\t\t\t\t\treturn res;\n"
        code += "\t\t\t\t} catch (...) {\n"
        code += "\t\t\t\t\treturn NULL;\n"
        code += "\t\t\t\t}\n"

        if len(fkt.signature) > 0:
            code += "\t\t\t} else\n"
            code += "\t\t\t\tPyErr_Clear();\n"

        code += "\t\t}\n"
        
    def stripArg(arg):
        if isinstance(arg, Object):
            delattr(arg, "parent")
            if hasattr(arg, "instance"):
                delattr(arg, "instance")
            for name, value in list(arg.attrs.items()):
                arg.attrs[name] = stripArg(value)
        else:
            if not arg.dtype in [bool, int, float]:
                arg.dtype = NPY_TYPE[arg.dtype]
            arg.dims = len(arg.shape)
            delattr(arg, "shape")
            if not isinstance(arg, ObjectAttr):
                delattr(arg, "scope")
                delattr(arg, "allocated")
            else:
                delattr(arg, "parent")
        return arg

    signatures = []
    for fkt in modtoken.functions[modtoken.main]:
        signatures.append([stripArg(copy.deepcopy(arg)) for arg in fkt.signature])

    if sys.version_info[0] == 2:
        pickled = pickle.dumps(signatures).replace("\n", "\\n")
    else:
        import base64
        pickled = base64.encodebytes(pickle.dumps(signatures)).decode('ascii').replace("\n", "\\n")

    code += "\t\tPyObject * signatures = Py_BuildValue(\"(sO)\", \"{0}\", args);\n".format(pickled)
    code += "\t\tif (!signatures) {\n"

    # TODO: make all exceptions reasonamble: http://docs.python.org/2/c-api/exceptions.html
    code += "\t\t\tPyErr_SetString(PyExc_ValueError, \"Error building signature string for {0}\");\n".format(modtoken.main)
    code += "\t\t\treturn NULL;\n"
    code += "\t\t}\n"
    code += "\t\treturn PyObject_Call(create_signature, signatures, NULL);\n"

    return code



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate the supported operators

http://pythonhosted.org/hope/lang.html#operators

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil
import numpy as np
import hope

prefix = "hope_operators"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def assign(d):
    a = d
    return a

@hope.jit
def invert(d):
    a = -d
    return a

@hope.jit
def add(a, b):
    return a + b

def example():
    #assigning
    d = 1 
    r = assign(d)
    assert r == d

    d = np.array([1])
    r = assign(d)
    assert np.all(r == d)
    
    d = np.array([[1], [2]])
    r = assign(d)
    assert np.all(r == d)

    #unary operation
    d = 1 
    r = invert(d)
    assert r == -d

    d = np.array([1])
    r = invert(d)
    assert np.all(r == -d)
    
    d = np.array([[1], [2]])
    r = invert(d)
    assert np.all(r == -d)
    
    #add (same for -,*, /)
    a, b = 1, 1 
    r = add(a, b)
    assert r == a + b

    a, b = np.array([1]), np.array([1])
    r = add(a, b)
    assert np.all(r == a + b)
    
    a, b = np.array([[1], [2]]), np.array([[1], [2]])
    r = add(a, b)
    assert np.all(r == a + b)

if __name__ == '__main__':
    example()


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test calls for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import hope
import itertools
import pytest

from .utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

def fkt_call_local_fun_callback(a, b):
    a[:] = b
@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_call_local_fun(dtype, shape):
    def fkt(a, b, c):
        fkt_call_local_fun_callback(a, b)
        c[:] = a
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

def fkt_call_local_fun_return_callback(a):
    return a
@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_call_local_fun_return(dtype, shape):
    def fkt(a):
        return fkt_call_local_fun_return_callback(a)
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)

@hope.jit
def fkt_call_jit_fun_return_callback(a):
    return a
@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_call_jit_fun_return(dtype, shape):
    def fkt(a):
        return fkt_call_jit_fun_return_callback(a)
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)

def fkt_call_scalar_fun_return_callback(a):
    return a + 1
@pytest.mark.parametrize("dtype", dtypes)
def test_call_scalar_fun_return(dtype):
    def fkt(a):
        return fkt_call_scalar_fun_return_callback(a)
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)

@hope.jit
def fkt_call_scalar_jit_fun_return_callback(a):
    return a + 1
@pytest.mark.parametrize("dtype", dtypes)
def test_call_scalar_jit_fun_return(dtype):
    def fkt(a):
        return fkt_call_scalar_jit_fun_return_callback(a)
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, []), random(dtype, [])
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)
    co, ch = fkt(ao), hfkt(ah)
    assert check(co, ch)

@hope.jit
def fkt_recursion_jit_callback(n):
    if n < 2:
        return n
    return fkt_recursion_jit_callback(n - 1) + fkt_recursion_jit_callback(n - 2)
def test_recursion_jit():
    assert fkt_recursion_jit_callback(20) == 6765

@hope.jit
def fkt_call_scalar_jit_multi_callback(a):
    return a + 1
def fkt_call_scalar_jit_multi_fkt(a):
    return fkt_call_scalar_jit_multi_callback(a)
@pytest.mark.parametrize("dtype", dtypes)
def test_fkt_call_scalar_jit_multi(dtype):
    hfkt = hope.jit(fkt_call_scalar_jit_multi_fkt)
    (a, _), (c, _) = random(dtype, []), random(dtype, [])
    c = hfkt(a)
    assert check(c, a + 1)
    c = hfkt(a)
    assert check(c, a + 1)
@pytest.mark.parametrize("dtype", dtypes)
def test_fkt_call_scalar_jit_multi_2(dtype):
    hfkt = hope.jit(fkt_call_scalar_jit_multi_fkt)
    (a, _), (c, _) = random(dtype, []), random(dtype, [])
    c = hfkt(a)
    assert check(c, a + 1)
    c = hfkt(a)
    assert check(c, a + 1)

def fkt_recursion_callback(n):
    if n < 2:
        return n
    return fkt_recursion_callback(n - 1) + fkt_recursion_callback(n - 2)
def test_recursion():
    hope_fkt_recursion_callback = hope.jit(fkt_recursion_callback)
    assert hope_fkt_recursion_callback(20) == 6765

# TODO: make tests for function calls with return types and expr / local vars as arguments



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from .utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module


np_version = tuple(map(int, np.__version__.split(".")))

#hope.config.keeptemp = True

def test_ufig_sin_cos():
    sinTable = np.sin(np.array(range(0, (1 << 11) + 1), dtype=np.float64) * 2. * np.pi / np.float64(1 << 11))
    cosTable = np.cos(np.array(range(0, (1 << 11) + 1), dtype=np.float64) * 2. * np.pi / np.float64(1 << 11))
    rngBuffer = np.random.randint(0, 1<<32, size=(44497,)).astype(np.uint32)

    def fkt_sncn(pt, rng, sinTable, cosTable):
        scL = rng >> (32 - 11)
        scB = np.float64(rng & np.uint32((1 << (32 - 11)) - 1)) / np.float64(1 << (32 - 11))
        scA = 1. - scB
        pt[0] = scA * sinTable[scL] + scB * sinTable[scL + 1]
        pt[1] = scA * cosTable[scL] + scB * cosTable[scL + 1]

    rngBuffer = np.random.randint(0, 1<<32, size=(1000,)).astype(np.uint32)
    pt, hpt = np.array([0., 0.]), np.array([0., 0.])
    hope.config.optimize = True
    hsncn = hope.jit(fkt_sncn)

    for rng in rngBuffer:
        fkt_sncn(pt, rng, sinTable, cosTable)
        hsncn(hpt, rng, sinTable, cosTable)
        assert check(pt, hpt)
        ang = rng / float(1 << 32) * 2. * np.pi
        assert np.all((hpt - np.array([np.sin(ang), np.cos(ang)])) / pt < 1e-5)
    hope.config.optimize = False

def test_ufig_gal_intrinsic():
    try:
        from scipy.special import gammaincinv
    except ImportError:
        pytest.skip("Scipy is not available")

    n = np.uint32(1.88 * np.float64(np.int64(1) << 32) / 10.)
    sersicLTable = np.uint32(n >> np.uint32(32 - 9))
    sersicBTable = np.float32(n - (sersicLTable << np.uint32(32 - 9))) / np.float32(1 << (32 - 9))

    # gamma lookup has 1<<9, 512 elements, and a more precise fitt on 0-1/(1<<3) with 1<<11, 2048 elements
    radiusTable = np.empty(((1 << 9) + 1, (1 << (11 + 1)) + 1, ), dtype=np.float32)
    radiusTable[0][0:(1 << 11)] = (np.power(gammaincinv(2e-15, np.float64(range(0, 1 << 11)) / np.float64(1 << 11)), 1e-15) / 1e-15).astype(np.float32)
    radiusTable[0][(1 << 11):(1 << (11 + 1))] = (np.power(gammaincinv(2e-15, 1. - 1. / np.float64(1 << 3) + np.float64(range(0, 1 << 11)) / np.float64(1 << (11 + 3))), 1e-15) / 1e-15).astype(np.float32)
    radiusTable[0][1 << 11] = (np.power(gammaincinv(2e-15, (1. - 1e-15) / np.float64(1 << 11)), 1e-15) / 1e-15).astype(np.float32)
    radiusTable[0][1 << (11 + 1)] = (np.power(gammaincinv(2e-15, 1. - 1. / np.float64(1 << 3) + (1. - 1e-15) / np.float64(1 << (11 + 3))), 1e-15) / 1e-15).astype(np.float32)

    # TODO: make only one gamma interpolation instead of two
    for i in range(1, (1 << 9) + 1):
        n = 10. * np.float64(i << (32 - 9)) / (np.int64(1) << 32)
        k = gammaincinv(2. * n, 0.5)
        radiusTable[i][0:(1 << 11)] = (np.power(gammaincinv(2. * n, np.float64(range(0, 1 << 11)) / np.float64(1 << 11)), n) / np.power(k, n)).astype(np.float32)
        radiusTable[i][(1 << 11):(1 << (11 + 1))] = (np.power(gammaincinv(2. * n, 1. - 1. / np.float64(1 << 3) + np.float64(range(0, 1 << 11)) / np.float64(1 << (11 + 3))), n) / np.power(k, n)).astype(np.float32)
        radiusTable[i][1 << 11] = (np.power(gammaincinv(2. * n, (1. - 1e-15) / np.float64(1 << 11)), n) / np.power(k, n)).astype(np.float32)
        radiusTable[i][1 << (11 + 1)] = (np.power(gammaincinv(2. * n, 1. - 1. / np.float64(1 << 3) + (1. - 1e-15) / np.float64(1 << (11 + 3))), n) / np.power(k, n)).astype(np.float32)

    def fkt_intrinsic(rng, sersicLTable, sersicBTable, radiusTable):
        drMaski = rng >> (32 - 3) == (1 << 3) - 1
        drKi = rng >> np.uint32(drMaski * 3)
        drLi = (drKi >> (32 - 11)) + np.uint32(drMaski * (1 << 11))
        drBi = np.float64(drKi & ((1 << (32 - 11)) - 1)) / np.float64(1 << (32 - 11))
        drAi = 1. - drBi

        nLi = sersicLTable
        nBi = sersicBTable
        nAi = 1 - nBi

        return drAi * (nAi * radiusTable[nLi, drLi] + nBi * radiusTable[nLi, drLi + 1]) \
           + drBi * (nAi * radiusTable[nLi + 1, drLi] + nBi * radiusTable[nLi + 1, drLi + 1])

    rngBuffer = np.random.randint(0, 1<<32, size=(1000,)).astype(np.uint32)
    hope.config.optimize = True
    hintrinsic = hope.jit(fkt_intrinsic)

    for rng in rngBuffer:
        dr = fkt_intrinsic(rng, sersicLTable, sersicBTable, radiusTable)
        hdr = hintrinsic(rng, sersicLTable, sersicBTable, radiusTable)
        assert check(dr, hdr)
    hope.config.optimize = False


def test_ufig_bincount():

    def fkt_bincount(buffer, x, y, size, size_x, size_y):
        for idx in range(size):
            if x[idx] >= 0 and x[idx] <= size_x and y[idx] >= 0 and y[idx] <= size_y:
                x_bin = np.uint64(x[idx])
                y_bin = np.uint64(y[idx])
                buffer[x_bin, y_bin] += 1.
    hope.config.optimize = True
    hbincount = hope.jit(fkt_bincount)

    x = np.random.uniform(-1, 6, size=(1000,))
    y = np.random.uniform(-1, 6, size=x.shape)
    buffer, hbuffer = np.zeros((5, 5)), np.zeros((5, 5))

    fkt_bincount(buffer, x, y, x.size, 5., 5.)
    hbincount(hbuffer, x, y, x.size, 5., 5.)
    hope.config.optimize = False
    assert check(buffer, hbuffer)

@pytest.mark.parametrize("dtype", [float, np.float32, np.float64])
def test_ufig_star(dtype):
    b = 3.5
    a = 1. / np.sqrt(2. ** (1. / (b - 1.)) - 1.)
    r50 = 2

    center = np.array([10.141, 10.414])
    dims = np.array([20, 20])
    # coefficients generated by http://keisan.casio.com/has10/SpecExec.cgi?id=system/2006/1280624821, 7th order
    x1D = np.array([ \
          0.5 - 0.9491079123427585245262 / 2 \
        , 0.5 - 0.7415311855993944398639 / 2 \
        , 0.5 - 0.4058451513773971669066 / 2 \
        , 0.5 \
        , 0.5 + 0.4058451513773971669066 / 2 \
        , 0.5 + 0.7415311855993944398639 / 2 \
        , 0.5 + 0.9491079123427585245262 / 2 \
    ], dtype=dtype)
    w1D = np.array([ \
          0.1294849661688696932706 / 2 \
        , 0.2797053914892766679015 / 2 \
        , 0.38183005050511894495 / 2 \
        , 0.4179591836734693877551 / 2 \
        , 0.38183005050511894495 / 2 \
        , 0.2797053914892766679015 / 2 \
        , 0.1294849661688696932706 / 2 \
    ], dtype=dtype)
    w2D = np.outer(w1D, w1D)
    def fkt_pdf(density, dims, center, w2D, r50, b, a):
        for x in range(dims[0]):
            for y in range(dims[1]):
                dr = np.sqrt((x - center[0]) ** 2 + (y - center[1]) ** 2)
                density[x, y] = np.sum(w2D * 2 * (b - 1) / (2 * np.pi * (r50 * a)**2) * (1 + (dr / (r50 * a))**2)**(-b))
        return density
    hope.config.optimize = True
    hpdf = hope.jit(fkt_pdf)

    density = np.zeros((dims[0], dims[1]), dtype=dtype)
    fkt_pdf(density, dims, center, w2D, r50, b, a)
    hdensity = np.zeros((dims[0], dims[1]), dtype=dtype)
    hpdf(hdensity, dims, center, w2D, r50, b, a)
    hope.config.optimize = False
#     print("going to sleep")
#     import time
#     time.sleep(10)
    assert check(density, hdensity)

    if np.all(hdensity == 1):
        print("asdf")
    else:
        print("else")



#!/usr/bin/env python

import os
import sys
from setuptools.command.test import test as TestCommand
from setuptools import find_packages

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ["-v", "-s"]
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

readme = open("README.rst").read()
history = open("HISTORY.rst").read().replace(".. :changelog:", "")
# during runtime
requires = ["numpy>=1.7", "sympy>=0.7.5"]
# for testing
tests_require = ["pytest>=2.3", "mock>=1.0.1", "tox>=1.8.0",
                 "coverage>=3.7.1", "pytest-cov>=1.8.0"]
# for development and documentation
extras_require = {
    "docs": "Sphinx>=1.2.3",
    "lint": "flake8>=2.2.4",
}

PACKAGE_PATH = os.path.abspath(os.path.join(__file__, os.pardir))

setup(
    name="hope",
    version="0.7.1",
    description="A Python Just-In-Time compiler for astrophysical computations",
    long_description=readme + "\n\n" + history,
    author="Lukas Gamper, Joel Akeret",
    author_email="hope@phys.ethz.ch",
    url="http://hope.phys.ethz.ch",
    packages=find_packages(PACKAGE_PATH, "test"),
    package_dir={"hope": "hope"},
    include_package_data=True,
    install_requires=requires,
    tests_require=tests_require,
    extra_requires=extras_require,
    license="GPLv3",
    zip_safe=False,
    keywords="HOPE, JIT compiler, HPC, high performance computing",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Compilers",
    ],
    cmdclass = { "test": PyTest },
)



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from .utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", [[np.int8, []]])
@make_test
def test_binary_and(a, b, c): return a == 1 and b == 1

@pytest.mark.parametrize("dtype,shape", [[np.int8, []]])
@make_test
def test_binary_or(a, b, c): return a == 1 or b == 1



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Test operators for `hope` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np
import hope, itertools, pytest, sys, sysconfig, os, shutil

from .utilities import random, check, make_test, JENKINS, min_dtypes, dtypes, shapes, setup_module, setup_method, teardown_module

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
@make_test
def test_unary_plus(a, b, c): c[:] = +a

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_binary_plus(dtype, shape):
    def fkt(a, b, c):
        c[:] = a + b
    hfkt = hope.jit(fkt)
    (ao, ah), (bo, bh), (co, ch) = random(dtype, shape), random(dtype, shape), random(dtype, shape)
    ao, ah, bo, bh = (ao / 2.).astype(dtype), (ah / 2.).astype(dtype), (bo / 2.).astype(dtype), (bh / 2.).astype(dtype)
    ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    ro, rh = fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtypea,dtypeb,dtypec", itertools.product(min_dtypes, min_dtypes, min_dtypes))
def test_cross_add_dtype(dtypea, dtypeb, dtypec):
    def fkt(a, b, c):
        c[:] = a + b
    hfkt = hope.jit(fkt)
    dtypeao = np.float64 if dtypea == float else dtypea
    dtypebo = np.float64 if dtypeb == float else dtypeb
    (ao, ah), (bo, bh), (co, ch) = random(dtypeao, [100]), random(dtypebo, [100]), random(dtypec, [100])
    ao, ah = (ao / 2.).astype(dtypea), (ah / 2.).astype(dtypea)
    bo, bh = (bo / 2.).astype(dtypeb), (bh / 2.).astype(dtypeb)
    fkt(ao.astype(dtypea), bo.astype(dtypeb), co),  hfkt(ah.astype(dtypea), bh.astype(dtypeb), ch)
    assert check(co, ch)
    fkt(ao.astype(dtypea), bo.astype(dtypeb), co),  hfkt(ah.astype(dtypea), bh.astype(dtypeb), ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype", dtypes)
def test_cross_add_scalar(dtype):
    def fkt(a, b, c):
        c[:] = a + b
    hfkt = hope.jit(fkt)
    dtypeo = np.float64 if dtype == float else dtype
    (ao, ah), (bo, bh), (co, ch) = random(dtypeo, [100]), random(dtypeo, []), random(dtypeo, [100])
    ao, ah = (ao / 2.).astype(dtype), (ah / 2.).astype(dtype)
    bo, bh = (bo / 2.).astype(dtype), (bh / 2.).astype(dtype)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_cross_add_shape(dtype, shape):
    def fkt(a, b, c):
        c[:] = a + b
    hfkt = hope.jit(fkt)
    dtypeo = np.float64 if dtype == float else dtype
    (ao, ah), (bo, bh), (co, ch) = random(dtypeo, []), random(dtypeo, shape), random(dtypeo, shape)
    ao, ah = (ao / 2.).astype(dtype), (ah / 2.).astype(dtype)
    bo, bh = (bo / 2.).astype(dtype), (bh / 2.).astype(dtype)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(ao, bo, co),  hfkt(ah, bh, ch)
    assert check(co, ch)
    fkt(bo, ao, co),  hfkt(bh, ah, ch)
    assert check(co, ch)
    fkt(bo, ao, co),  hfkt(bh, ah, ch)
    assert check(co, ch)

@pytest.mark.parametrize("dtype,shape", itertools.product(dtypes, shapes[1:]))
def test_augmented_plus(dtype, shape):
    def fkt(a, c):
        c[:] += a
    hfkt = hope.jit(fkt)
    (ao, ah), (co, ch) = random(dtype, shape), random(dtype, shape)
    ao, ah, co, ch = (ao / 4.).astype(dtype), (ah / 4.).astype(dtype), (co / 2.).astype(dtype), (ch / 2.).astype(dtype)
    ro, rh = fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)
    ro, rh = fkt(ao, co),  hfkt(ah, ch)
    assert check(co, ch)



# Copyright (c) 2013 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

from __future__ import print_function, division, absolute_import, unicode_literals

import sys

from hope.options import get_cxxflags

# Additional compiler flags, formated as array of strings
cxxflags = get_cxxflags()
"""
List of c++ compiler flags. Normally hope does determing the right flags itself.
"""

prefix = ".hope.{}.{}".format(*sys.version_info[:2])
"""
Prefix of the folder hope saves all data in.
"""

del sys

verbose = False
"""
Print a intermediate representation of each function during compilation.
"""

optimize = False
"""
Use '''sympy''' to simplify expression and exptract common subexpression detection
"""

keeptemp = False
"""
Keep the intermediate c++ source and compiler output generated during compilation.
"""

rangecheck = False
"""
Check if indeces are out of bounds
"""

hopeless = False
"""
Disable hope. If hope.config.hopeless is True, hope.jit return the original function.
Use this function for debug purpos
"""

# make readable cpp file, but typecasting is not exactly the same as in numpy - this flag is private
_readablecxx = False



# Copyright (c) 2014 ETH Zurich, Institute of Astronomy, Lukas Gamper <lukas.gamper@usystems.ch>

# TODO: remove comment!
# replace pow(a, number(int)) by multiplications
# find subexpressions and factor them out => make this blockwise (for that merge 1d expressions into one block)

# 1: merge subexpressions
# 2: create blacklist out of find altered values 
# 3: factor out subexpressions

from __future__ import print_function, division, absolute_import, unicode_literals


from hope._ast import *
from hope._const import *
from hope._dump import Dumper

import sympy as sp

# http://docs.sympy.org/latest/modules/functions/elementary.html
SYM_UNARY_OPERATORS = {}
SYM_UNARY_OPERATORS["+"] = lambda a: +a
SYM_UNARY_OPERATORS["-"] = lambda a: -a

SYM_BINARY_OPERATORS = {}
SYM_BINARY_OPERATORS["+"] = lambda a, b: a + b
SYM_BINARY_OPERATORS["-"] = lambda a, b: a - b
SYM_BINARY_OPERATORS["*"] = lambda a, b: a * b
SYM_BINARY_OPERATORS["/"] = lambda a, b: a / b
SYM_BINARY_OPERATORS["**"] = lambda a, b: a ** b
# SYM_BINARY_OPERATORS["%"] = lambda a, b: a % b
# SYM_BINARY_OPERATORS["<<"] = lambda a, b: a << b
# SYM_BINARY_OPERATORS[">>"] = lambda a, b: a >> b
# SYM_BINARY_OPERATORS["|"] = lambda a, b: a | b
# SYM_BINARY_OPERATORS["^"] = lambda a, b: a ^ b
# SYM_BINARY_OPERATORS["&"] = lambda a, b: a & b

SYM_COMPARE_OPERATORS = {}
SYM_COMPARE_OPERATORS["=="] = lambda a, b: a == b
SYM_COMPARE_OPERATORS["!="] = lambda a, b: a != b
SYM_COMPARE_OPERATORS["<"] = lambda a, b: a < b
SYM_COMPARE_OPERATORS["<="] = lambda a, b: a <= b
SYM_COMPARE_OPERATORS[">"] = lambda a, b: a > b
SYM_COMPARE_OPERATORS[">="] = lambda a, b: a >= b

SYM_NUMBER = {}
SYM_NUMBER[bool] = lambda a: sp.BooleanTrue if a else sp.BooleanFalse
SYM_NUMBER[int] = lambda a: sp.Integer(a)
SYM_NUMBER[float] = lambda a: sp.Float(a, -np.log10(np.finfo(np.float_).resolution))

SYM_UNARY_FUNCTIONS = {}
SYM_UNARY_FUNCTIONS["sin"] = sp.sin
SYM_UNARY_FUNCTIONS["cos"] = sp.cos
SYM_UNARY_FUNCTIONS["tan"] = sp.tan

class CreateExprVisitor(NodeVisitor):
    def __init__(self, dumper): 
        self.dumper = dumper
    def visit_Number(self, node):
        return SYM_NUMBER[node.dtype](node.value)
    def visit_Variable(self, node):
        if not self.dumper.visit(node) in self.symbols:
            raise Exception("Unknown expression: {0!s}".format(self.dumper.visit(node)))
        return sp.Symbol(self.dumper.visit(node))
    def visit_Object(self, node):
        raise Exception("Unable to create expression from : {0!s}".format(self.dumper.visit(node)))
    def visit_ObjectAttr(self, node):
        if not self.dumper.visit(node) in self.symbols:
            raise Exception("Unknown expression: {0!s}".format(self.dumper.visit(node)))
        return sp.Symbol(self.dumper.visit(node))
    def visit_View(self, node):
        if not self.dumper.visit(node.variable) in self.symbols:
            raise Exception("Unknown expression: {0!s}".format(self.dumper.visit(node)))
        self.symbols[self.dumper.visit(node)] = node
        return sp.Symbol(self.dumper.visit(node))
    def visit_UnaryOp(self, node):
        return SYM_UNARY_OPERATORS[node.op](self.visit(node.operand))
    def visit_BinOp(self, node):
        return SYM_BINARY_OPERATORS[node.op](self.visit(node.left), self.visit(node.right))
    def visit_Compare(self, node):
        return SYM_COMPARE_OPERATORS[node.op](self.visit(node.left), self.visit(node.comparator))
    def visit_Call(self, node):
        return self.visit(node.name)(*[self.visit(arg) for arg in node.args])
    # TODO: implement this!
    # def visit_GlobalFunction(self, node): ???
    # TODO: implement this!
    # def visit_HopeAttr(self, node): ???
    def visit_NumpyAttr(self, node):
        if node.name in SYM_UNARY_FUNCTIONS:
            return SYM_UNARY_FUNCTIONS[node.name]
        else:
            return sp.Function('np.' + node.name)
    # TODO: implement this!
    # def visit_NumpyContraction(self, node): ???

class CheckOptimizeVisitor(NodeVisitor):

    def visit_Number(self, node): return True
    def visit_Variable(self, node): return True
    def visit_Object(self, node): return True
    def visit_ObjectAttr(self, node): return True
    # TODO: can wie optimize the index in the extent?
    def visit_View(self, node): return True
    def visit_Expr(self, node): return False
    def visit_Assign(self, node): return False
    def visit_AugAssign(self, node): return False
    def visit_UnaryOp(self, node): return self.visit(node.operand)
    def visit_BinOp(self, node): 
        return self.visit(node.left) and self.visit(node.right) if node.op in SYM_BINARY_OPERATORS else False
    def visit_BoolOp(self, node): return False
    def visit_Compare(self, node): return self.visit(node.left) and self.visit(node.comparator)

    def visit_If(self, node): return False
    def visit_For(self, node): return False
    def visit_While(self, node): return False

    def visit_Call(self, node):
        return self.visit(node.name) and np.all([self.visit(value) for value in list(node.keywords.values())]
                                                + [self.visit(arg) for arg in node.args])

    def visit_GlobalFunction(self, node): return True
    def visit_HopeAttr(self, node): return False

    def visit_NumpyAttr(self, node): 
        return node.name in SYM_UNARY_FUNCTIONS

    def visit_NumpyContraction(self, node): return False
    def visit_Allocate(self, node): return False
    def visit_Return(self, node): return False
    def visit_Block(self, node): return False
    def visit_Body(self, node): return False

class SympyPowVisitor(object):
    def visit(self, expr):
        if expr.is_Pow and expr.exp.is_Integer:
            return [expr]
        else:
            return [item for arg in expr.args for item in self.visit(arg)]

class SympyToAstVisitor(object):

    def __init__(self):
        for name in list(SYM_UNARY_FUNCTIONS.keys()):
            if not name in NPY_UNARY_FUNCTIONS and not name in NPY_CAST_FUNCTIONS:
                raise Exception("Unknown Function {0}".format(name))
            setattr(self, "visit_{0}".format(name), self.npUnaryFunction_visit)

    def visit(self, expr):
        if hasattr(self, 'visit_{0}'.format(type(expr).__name__)):
            return getattr(self, 'visit_{0}'.format(type(expr).__name__))(expr)
        else:
            return self.generic_visit(expr)

    def generic_visit(self, expr):
        raise Exception("Not Implemented Expression: {0}: {1!s}".format(type(expr).__name__, expr))

    def visit_Add(self, expr):
        ret = self.visit(expr.args[0])
        for term in expr.args[1:]:
            node = self.visit(term)
            if isinstance(node, UnaryOp) and node.op == UNARY_OPERATORS["USub"]:
                ret = BinOp("Sub", ret, node.operand)
            else:
                ret = BinOp("Add", ret, node)
        return ret

    def visit_Mul(self, expr):
        sign, numerators, denominator = 1, [], []
        for term in expr.as_ordered_factors():
            if term.is_Pow and term.exp.is_Rational and term.exp.is_negative:
                denominator.append(sp.Pow(term.base, -term.exp, evaluate=term.exp==-1))
            elif term.is_Rational:
                if term.p == -1:
                    sign *= -1
                elif term.p != 1:
                    numerators.append(sp.Rational(term.p))
                if term.q == -1:
                    sign *= -1
                elif term.q != 1:
                    denominator.append(sp.Rational(term.q))
            else:
                numerators.append(term)
        if len(numerators) == 0:
            ret = Number(1)
        else:
            ret = self.visit(numerators[0])
            for arg in numerators[1:]:
                ret = BinOp("Mult", ret, self.visit(arg))
        if len(denominator) > 0:
            ret = BinOp("Div", ret, self.binOp_visit("Mult", denominator))
        if sign < 0:
            ret = UnaryOp("USub", ret)
        return ret

    def visit_Pow(self, expr):
        from sympy.core.singleton import S
        if expr.exp.is_real and float(int(expr.exp)) == expr.exp:
            print("Integer exponent as flaot: {0!s}".format(expr))
        if expr.exp is S.Half or (expr.exp.is_real and expr.exp == 0.5):
            return Call(NumpyAttr("sqrt"), True, [self.visit(expr.base)])
        elif -expr.exp is S.Half or (expr.exp.is_real and -expr.exp == 0.5):
            return BinOp("Div", Number(1.), Call(NumpyAttr("sqrt"), True, [self.visit(expr.base)], {}))
        elif expr.exp == -1:
            return BinOp("Div", Number(1.), self.visit(expr.base))
        else:
            return self.binOp_visit("Pow", expr.args)

    def visit_Integer(self, expr):
        return Number(expr.p)
    def visit_Rational(self, expr):
        return BinOp("Div", Number(float(expr.p)), Number(float(expr.q)))
    def visit_Float(self, expr):
        return Number(float(expr))
    def visit_Symbol(self, expr):
        if not expr.name in self.symbols:
            raise Exception("Unknown symbol: {0!s}".format(expr.name))
        return self.symbols[expr.name]
    def visit_NegativeOne(self, expr):
        return Number(-1)
    def visit_Zero(self, expr):
        return Number(0)
    def visit_One(self, expr):
        return Number(1)

    def npUnaryFunction_visit(self, expr):
        return Call(NumpyAttr(type(expr).__name__), True, [self.visit(arg) for arg in expr.args], {})
    def binOp_visit(self, name, args):
        ret = self.visit(args[0])
        for arg in args[1:]:
            ret = BinOp(name, ret, self.visit(arg))
        return ret

class Optimizer(NodeVisitor):
    def __init__(self):
        self.dumper = Dumper()
        self.checkVisitor, self.createExpr, self.sympyToAst, self.sympyPow, self.next = CheckOptimizeVisitor(), CreateExprVisitor(self.dumper), SympyToAstVisitor(), SympyPowVisitor(), 0

    def visit_Number(self, node): pass
    def visit_NewVariable(self, node): pass
    def visit_Variable(self, node): pass
    def visit_Object(self, node): pass
    def visit_ObjectAttr(self, node): pass
    def visit_Dimension(self, node): pass
    def visit_View(self, node): pass
    def visit_Expr(self, node): pass
    def visit_Assign(self, node): pass
    def visit_AugAssign(self, node): pass
    def visit_UnaryOp(self, node): pass
    def visit_BinOp(self, node): pass
    def visit_Compare(self, node): pass

    def visit_If(self, node):
        # TODO: optimize condition
        # if condition is compile time -> remove!
        self.visit(node.body)
        if not node.orelse is None:
            self.visit(node.orelse)

    def visit_For(self, node):
        self.symbols[self.dumper.visit(node.iter)] = node.iter
        self.visit(node.body)
        del self.symbols[self.dumper.visit(node.iter)]
    def visit_While(self, node):
        self.visit(node.body)

    def visit_Call(self, node): pass
    def visit_GlobalFunction(self, node): pass
    def visit_HopeAttr(self, node): pass
    def visit_NumpyAttr(self, node): pass
    def visit_NumpyContraction(self, node): pass

    def visit_Allocate(self, node):
        self.symbols[self.dumper.visit(node.variable)] = node.variable

    def visit_Return(self, node): pass

    def visit_Block(self, node):
        body, knownexprs, powexprs = [], {}, {}
        for astexpr in node.body:
            self.visit(astexpr)
            if isinstance(astexpr, Assign):
                if isinstance(astexpr.target, View):
                    self.symbols[self.dumper.visit(astexpr.target.variable)] = astexpr.target
                elif isinstance(astexpr.target, Variable):
                    self.symbols[self.dumper.visit(astexpr.target)] = astexpr.target
                else:
                    raise Exception("Unknown token".format(self.dumper.visit(astexpr.target)))
            # TODO: implement for expr
            # TODO: replace subexpressions over several lines
            if isinstance(astexpr, (Assign, AugAssign)) and self.checkVisitor.visit(astexpr.value):
                symexpr = sp.simplify(self.createExpr.visit(astexpr.value))
                subexprs, newexprs = sp.cse(symexpr, optimizations='basic')
                if len(newexprs) != 1:
                    raise Exception("Error running Common Subexpression Detection for {1!s}".format(symexpr))
                newexpr = newexprs[0]
                for symbol, subexpr in subexprs:
                    for subsymbol, newsymbol in list(knownexprs.items()):
                        subexpr = subexpr.subs(subsymbol, newsymbol)
                    for powexpr in self.sympyPow.visit(subexpr):
                        subexpr, _ = self.replace_pow(body, subexpr, powexprs, powexpr, np.abs(powexpr.exp.p))
                    value = self.sympyToAst.visit(sp.simplify(subexpr))
                    name, self.next = "__sp{0}".format(self.next), self.next + 1
                    self.symbols[name] = Variable(name, copy.deepcopy(value.shape), value.dtype)
                    body.append(Assign(self.symbols[name], value))
                    knownexprs[symbol] = sp.Symbol(name)
                    newexpr = newexpr.subs(symbol, knownexprs[symbol])
                for powexpr in sorted(self.sympyPow.visit(newexpr), key=lambda x: -np.abs(x.exp.p)):
                    newexpr, _ = self.replace_pow(body, newexpr, powexprs, powexpr, np.abs(powexpr.exp.p))
                newvalue = self.sympyToAst.visit(sp.simplify(newexpr))
                if astexpr.value.dtype != newvalue.dtype:
                    if isinstance(newvalue, Number):
                        newvalue = Number(astexpr.value.dtype(newvalue.value))
                    else:
                        raise Exception("dtype does not match {0} != {1}".format(self.dumper.visit(astexpr.value), self.dumper.visit(newvalue)))
                if not(len(astexpr.target.shape) > 0 and len(newvalue.shape) == 0):
                    if len(astexpr.value.shape) != len(newvalue.shape):
                        raise Exception("length of shape does not match {0} != {1}".format(self.dumper.visit(astexpr.value), self.dumper.visit(newvalue)))
                    for extent1, extent2 in zip(astexpr.value.shape, newvalue.shape):
                        (lower1, upper1), (lower2, upper2) = extent1, extent2
                        if not ((lower1 is None and lower2 is None) or lower1 == lower2) or upper1 != upper2:
                            raise Exception("shape does not match {0} != {1}".format(self.dumper.visit(astexpr.value), self.dumper.visit(newvalue)))
                astexpr.value = newvalue
                body.append(astexpr)
            else:
                body.append(astexpr)
        node.body = body

    def visit_Body(self, node):
        for block in node.blocks:
            self.visit(block)

    def visit_FunctionDef(self, node):
        if not node.optimized:
            self.symbols = {}
            for var in node.signature:
                self.add_symbol(var)
            self.createExpr.symbols, self.sympyToAst.symbols = self.symbols, self.symbols
            node.optimized = True
            self.visit(node.body)

    def visit_Module(self, node):
        for fktcls in list(node.functions.values()):
            for fkt in fktcls:
                self.visit(fkt)

    def add_symbol(self, symbol):
        if isinstance(symbol, Object):
            for attr in list(symbol.attrs.values()):
                self.add_symbol(attr)
        else:
            self.symbols[self.dumper.visit(symbol)] = symbol

    def replace_pow(self, body, symexpr, powexprs, expr, exp):
        if exp == 1:
            return (symexpr, None)
        elif not (expr.base, exp) in powexprs:
            if exp == 2:
                operand = sp.simplify(expr.base)
                value = BinOp("Mult", self.sympyToAst.visit(operand), self.sympyToAst.visit(operand))
            elif exp % 2 == 1:
                _, operand = self.replace_pow(body, symexpr, powexprs, expr, exp - 1)
                value = BinOp("Mult", self.symbols[operand], self.sympyToAst.visit(sp.simplify(expr.base)))
            else:
                _, operand = self.replace_pow(body, symexpr, powexprs, expr, exp / 2)
                value = BinOp("Mult", self.symbols[operand], self.symbols[operand])
            name, self.next = "__sp{0}".format(self.next), self.next + 1
            self.symbols[name] = Variable(name, copy.deepcopy(value.shape), value.dtype)
            body.append(Assign(self.symbols[name], value))
            powexprs[(expr.base, exp)] = name
        if np.abs(expr.exp.p) == exp:
            symbol = sp.Symbol(powexprs[(expr.base, exp)])
            symexpr = symexpr.subs(expr, self.symbols[powexprs[(expr.base, exp)]].dtype(1) / symbol if expr.exp.is_negative else symbol)
        return (symexpr, powexprs[(expr.base, exp)])



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
A simple example to demonstrate if-elif-else support

http://pythonhosted.org/hope/lang.html#conditional-expressions

Created on Oct 23, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import os
import shutil

import hope

prefix = "hope_conditions"

hope.config.keeptemp = True
hope.config.verbose = True
hope.config.prefix = prefix

if os.path.exists(prefix):
    shutil.rmtree(prefix)

@hope.jit
def demo(d):
    if d == 0:
        return d + 1
    elif d == 1:
        return d
    else:
        return d -1 

def example():
    r = demo(0)
    assert r == 1

    r = demo(1)
    assert r == 1

    r = demo(2)
    assert r == 1

if __name__ == '__main__':
    example()